#!/usr/bin/env python3
import logging
import asyncio
import random

from bitarray import bitarray
from bitarray.util import int2ba, ba2int

import ebyst

from ebyst.interfaces import MT25Q, MDIO, DDR3, DDR4
from ebyst import Pin, PinGroup, DiffPin

logger = logging.getLogger(__name__)


async def leds(ctl, dev):
    LEDS = ['IO_H21', 'IO_H22', 'IO_F23', 'IO_C27', 'IO_D25', 'IO_C26', 'IO_B26', 'IO_F22']
    for i in range(10):
        for pin in LEDS + LEDS[::-1]:
            dev.pinmap[pin].output_enable()
            dev.pinmap[pin].set_value(1)
            await ctl.cycle()
            dev.pinmap[pin].set_value(0)

async def flash(ctl, dev):
    pins = {
        'C':        dev.pinmap["IO_H19"],
        'Sn':       dev.pinmap["IO_A20"],
        'RESETn':   dev.pinmap["IO_A24"],
        'WPn':      dev.pinmap["IO_A27"],
        'HOLDn':    dev.pinmap["IO_A22"],
        'DQ0':      dev.pinmap["IO_G19"],
        'DQ1':      dev.pinmap["IO_C18"],
    }
    ctl.trace("flash.vcd", **pins)
    flash = MT25Q(ctl, **pins)

    await flash.init()
    print("Flash ID:", (await flash.read_id()).hex())

async def mdio(ctl, dev):
    pins = {
        'MDC':      dev.pinmap["IO_Y12"],
        'MDIO':     dev.pinmap["IO_Y13"],
    }
    ctl.trace("mdio.vcd", **pins)
    mdio = MDIO(ctl, **pins)

    # PHY ID
    for pin in ("IO_Y7", "IO_AA7", "IO_AA9"):
        dev.pinmap[pin].output_enable()
        dev.pinmap[pin].set_value(0)

    # PHY RESET
    dev.pinmap["IO_U11"].output_enable()
    dev.pinmap["IO_U11"].set_value(0)
    await ctl.cycle()
    dev.pinmap["IO_U11"].set_value(1)
    await ctl.cycle()

    await mdio.init()
    print("PHY ID: 0x%04x %04x" % (await mdio.read(0, 2), await mdio.read(0, 3)))

async def ddr3(ctl, dev):
    pins = {
        'DQ':       PinGroup([dev.pinmap["IO_AN22"], dev.pinmap["IO_AM24"], dev.pinmap["IO_AN21"], dev.pinmap["IO_AN24"],
                              dev.pinmap["IO_AP20"], dev.pinmap["IO_AP19"], dev.pinmap["IO_AP21"], dev.pinmap["IO_AN19"]]),
        'DQS':      DiffPin(dev.pinmap["IO_AP23"], dev.pinmap["IO_AP24"]),
        'DM':       dev.pinmap["IO_AN23"],
        'A':        PinGroup([dev.pinmap["IO_AL27"], dev.pinmap["IO_AL26"], dev.pinmap["IO_AM27"], dev.pinmap["IO_AN27"],
                              dev.pinmap["IO_AN26"], dev.pinmap["IO_AP25"], dev.pinmap["IO_AL25"], dev.pinmap["IO_AK25"],
                              dev.pinmap["IO_AJ23"], dev.pinmap["IO_AH23"], dev.pinmap["IO_AJ25"], dev.pinmap["IO_AJ24"],
                              dev.pinmap["IO_AL22"], dev.pinmap["IO_AK23"], dev.pinmap["IO_AL24"], dev.pinmap["IO_AL23"]]),
        'BA':       PinGroup([dev.pinmap["IO_AE25"], dev.pinmap["IO_AD23"], dev.pinmap["IO_AD25"]]),
        'ODT':      PinGroup([dev.pinmap["IO_AF23"], dev.pinmap["IO_AH25"]]),
        'CK':       DiffPin(dev.pinmap["IO_AP26"], dev.pinmap["IO_AP27"]),
        'CKE':      PinGroup([dev.pinmap["IO_AF22"], dev.pinmap["IO_AD24"]]),
        'CSn':      PinGroup([dev.pinmap["IO_AE22"], dev.pinmap["IO_AH24"]]),
        'RASn':     dev.pinmap["IO_AE23"],
        'CASn':     dev.pinmap["IO_AF25"],
        'WEn':      dev.pinmap["IO_AF24"],
        'RESETn':   dev.pinmap["IO_AG22"],
    }
    ctl.trace("ddr3.vcd", **pins)
    ddr3 = DDR3(ctl, **pins)

    bank = bitarray("000", endian='little')
    row = bitarray("0000000000000000", endian='little')
    column = bitarray("0000000000000000", endian='little')

    await ddr3.init()

    await ddr3.activate(ba=bank, ra=row)
    wdata = []
    for i in range(8):
        wdata.append(int2ba(random.randint(0, 255), 8, endian='little'))
    print(f"DDR3: Writing: {' '.join(['%02x' % ba2int(x) for x in wdata])}")
    await ddr3.write(ba=bank, ca=column, data=wdata)
    await ddr3.precharge(ba=bank)

    await ddr3.activate(ba=bank, ra=row)
    rdata = await ddr3.read(ba=bank, ca=column)
    await ddr3.precharge(ba=bank)
    print(f"DDR3: Read back: {' '.join(['%02x' % ba2int(x) for x in rdata])}")
    if rdata == wdata:
        print(f"DDR3: OK")
    else:
        print(f"DDR3: FAILED")

async def ddr4(ctl, dev):
    pins_a = {
        'DQ':       PinGroup([dev.pinmap["IO_AG10"], dev.pinmap["IO_AE12"], dev.pinmap["IO_AF10"], dev.pinmap["IO_AF12"],
                              dev.pinmap["IO_AD10"], dev.pinmap["IO_AD13"], dev.pinmap["IO_AE10"], dev.pinmap["IO_AD14"]]),
        'DQS':      DiffPin(dev.pinmap["IO_AE13"], dev.pinmap["IO_AF13"]),
        'DM':       dev.pinmap["IO_AE11"],
        'A':        PinGroup([dev.pinmap["IO_AE5"], dev.pinmap["IO_AF5"], dev.pinmap["IO_AF2"], dev.pinmap["IO_AF3"],
                              dev.pinmap["IO_AH3"], dev.pinmap["IO_AJ3"], dev.pinmap["IO_AH1"], dev.pinmap["IO_AH2"],
                              dev.pinmap["IO_AK3"], dev.pinmap["IO_AL3"], dev.pinmap["IO_AJ1"], dev.pinmap["IO_AK1"],
                              dev.pinmap["IO_AH4"], dev.pinmap["IO_AJ4"]]),
        'BA':       PinGroup([dev.pinmap["IO_AD8"], dev.pinmap["IO_AD6"]]),
        'BG':       PinGroup([dev.pinmap["IO_AG7"], dev.pinmap["IO_AG6"]]),
        'ODT':      PinGroup([dev.pinmap["IO_AE7"], dev.pinmap["IO_AF4"]]),
        'CK':       DiffPin(dev.pinmap["IO_AG4"], dev.pinmap["IO_AG5"]),
        'CKE':      PinGroup([dev.pinmap["IO_AG9"], dev.pinmap["IO_AF8"]]),
        'CSn':      PinGroup([dev.pinmap["IO_AF9"], dev.pinmap["IO_AE3"]]),
        'RASn':     dev.pinmap["IO_AD9"],
        'CASn':     dev.pinmap["IO_AL2"],
        'WEn':      dev.pinmap["IO_AK2"],
        'ACTn':     dev.pinmap["IO_AE6"],
        'ALERTn':   dev.pinmap["IO_AG2"],
        'PARITY':   dev.pinmap["IO_AE8"],
        'TEN':      dev.pinmap["IO_AG1"],
        'RESETn':   dev.pinmap["IO_AF7"],
    }
    pins_b = {
        'DQ':       PinGroup([dev.pinmap["IO_AH7"], dev.pinmap["IO_AJ9"], dev.pinmap["IO_AH6"], dev.pinmap["IO_AJ8"],
                              dev.pinmap["IO_AH9"], dev.pinmap["IO_AK7"], dev.pinmap["IO_AK5"], dev.pinmap["IO_AK6"]]),
        'DQS':      DiffPin(dev.pinmap["IO_AJ5"], dev.pinmap["IO_AJ6"]),
        'DM':       dev.pinmap["IO_AK8"],
        'A':        PinGroup([dev.pinmap["IO_AE5"], dev.pinmap["IO_AF5"], dev.pinmap["IO_AF2"], dev.pinmap["IO_AF3"],
                              dev.pinmap["IO_AH3"], dev.pinmap["IO_AJ3"], dev.pinmap["IO_AH1"], dev.pinmap["IO_AH2"],
                              dev.pinmap["IO_AK3"], dev.pinmap["IO_AL3"], dev.pinmap["IO_AJ1"], dev.pinmap["IO_AK1"],
                              dev.pinmap["IO_AH4"], dev.pinmap["IO_AJ4"]]),
        'BA':       PinGroup([dev.pinmap["IO_AD8"], dev.pinmap["IO_AD6"]]),
        'BG':       PinGroup([dev.pinmap["IO_AG7"], dev.pinmap["IO_AG6"]]),
        'ODT':      PinGroup([dev.pinmap["IO_AE7"], dev.pinmap["IO_AF4"]]),
        'CK':       DiffPin(dev.pinmap["IO_AG4"], dev.pinmap["IO_AG5"]),
        'CKE':      PinGroup([dev.pinmap["IO_AG9"], dev.pinmap["IO_AF8"]]),
        'CSn':      PinGroup([dev.pinmap["IO_AF9"], dev.pinmap["IO_AE3"]]),
        'RASn':     dev.pinmap["IO_AD9"],
        'CASn':     dev.pinmap["IO_AL2"],
        'WEn':      dev.pinmap["IO_AK2"],
        'ACTn':     dev.pinmap["IO_AE6"],
        'ALERTn':   dev.pinmap["IO_AG2"],
        'PARITY':   dev.pinmap["IO_AE8"],
        'TEN':      dev.pinmap["IO_AG1"],
        'RESETn':   dev.pinmap["IO_AF7"],
    }
    pins_c = {
        'DQ':       PinGroup([dev.pinmap["IO_AM1"], dev.pinmap["IO_AN2"], dev.pinmap["IO_AM2"], dev.pinmap["IO_AN1"],
                              dev.pinmap["IO_AM5"], dev.pinmap["IO_AP2"], dev.pinmap["IO_AL5"], dev.pinmap["IO_AP3"]]),
        'DQS':      DiffPin(dev.pinmap["IO_AL4"], dev.pinmap["IO_AM4"]),
        'DM':       dev.pinmap["IO_AN3"],
        'A':        PinGroup([dev.pinmap["IO_AE5"], dev.pinmap["IO_AF5"], dev.pinmap["IO_AF2"], dev.pinmap["IO_AF3"],
                              dev.pinmap["IO_AH3"], dev.pinmap["IO_AJ3"], dev.pinmap["IO_AH1"], dev.pinmap["IO_AH2"],
                              dev.pinmap["IO_AK3"], dev.pinmap["IO_AL3"], dev.pinmap["IO_AJ1"], dev.pinmap["IO_AK1"],
                              dev.pinmap["IO_AH4"], dev.pinmap["IO_AJ4"]]),
        'BA':       PinGroup([dev.pinmap["IO_AD8"], dev.pinmap["IO_AD6"]]),
        'BG':       PinGroup([dev.pinmap["IO_AG7"], dev.pinmap["IO_AG6"]]),
        'ODT':      PinGroup([dev.pinmap["IO_AE7"], dev.pinmap["IO_AF4"]]),
        'CK':       DiffPin(dev.pinmap["IO_AG4"], dev.pinmap["IO_AG5"]),
        'CKE':      PinGroup([dev.pinmap["IO_AG9"], dev.pinmap["IO_AF8"]]),
        'CSn':      PinGroup([dev.pinmap["IO_AF9"], dev.pinmap["IO_AE3"]]),
        'RASn':     dev.pinmap["IO_AD9"],
        'CASn':     dev.pinmap["IO_AL2"],
        'WEn':      dev.pinmap["IO_AK2"],
        'ACTn':     dev.pinmap["IO_AE6"],
        'ALERTn':   dev.pinmap["IO_AG2"],
        'PARITY':   dev.pinmap["IO_AE8"],
        'TEN':      dev.pinmap["IO_AG1"],
        'RESETn':   dev.pinmap["IO_AF7"],
    }
    pins_d = {
        'DQ':       PinGroup([dev.pinmap["IO_AH11"], dev.pinmap["IO_AG12"], dev.pinmap["IO_AG11"], dev.pinmap["IO_AH12"],
                              dev.pinmap["IO_AJ10"], dev.pinmap["IO_AJ14"], dev.pinmap["IO_AJ11"], dev.pinmap["IO_AJ13"]]),
        'DQS':      DiffPin(dev.pinmap["IO_AH14"], dev.pinmap["IO_AH13"]),
        'DM':       dev.pinmap["IO_AK13"],
        'A':        PinGroup([dev.pinmap["IO_AE5"], dev.pinmap["IO_AF5"], dev.pinmap["IO_AF2"], dev.pinmap["IO_AF3"],
                              dev.pinmap["IO_AH3"], dev.pinmap["IO_AJ3"], dev.pinmap["IO_AH1"], dev.pinmap["IO_AH2"],
                              dev.pinmap["IO_AK3"], dev.pinmap["IO_AL3"], dev.pinmap["IO_AJ1"], dev.pinmap["IO_AK1"],
                              dev.pinmap["IO_AH4"], dev.pinmap["IO_AJ4"]]),
        'BA':       PinGroup([dev.pinmap["IO_AD8"], dev.pinmap["IO_AD6"]]),
        'BG':       PinGroup([dev.pinmap["IO_AG7"], dev.pinmap["IO_AG6"]]),
        'ODT':      PinGroup([dev.pinmap["IO_AE7"], dev.pinmap["IO_AF4"]]),
        'CK':       DiffPin(dev.pinmap["IO_AG4"], dev.pinmap["IO_AG5"]),
        'CKE':      PinGroup([dev.pinmap["IO_AG9"], dev.pinmap["IO_AF8"]]),
        'CSn':      PinGroup([dev.pinmap["IO_AF9"], dev.pinmap["IO_AE3"]]),
        'RASn':     dev.pinmap["IO_AD9"],
        'CASn':     dev.pinmap["IO_AL2"],
        'WEn':      dev.pinmap["IO_AK2"],
        'ACTn':     dev.pinmap["IO_AE6"],
        'ALERTn':   dev.pinmap["IO_AG2"],
        'PARITY':   dev.pinmap["IO_AE8"],
        'TEN':      dev.pinmap["IO_AG1"],
        'RESETn':   dev.pinmap["IO_AF7"],
    }

    ctl.trace("ddr4.vcd", trace_all=True, **pins_a)

    for pins in (pins_a, pins_b, pins_c, pins_d):
        ddr4 = DDR4(ctl, **pins)
        await ddr4.init()
        print("DDR4: Running connectivity test")
        await ddr4.test()
        print("DDR4: Done")

async def clock(ctl, dev):
    CLK50 = dev.pinmap["IO_E25"]
    CLK50.output_enable(False)
    ctl.trace("clock.vcd", trace_all=True, CLK50=CLK50)
    ones = zeroes = 0
    for _ in range(100):
        await ctl.cycle()
        if CLK50.get_value():
            ones += 1
        else:
            zeroes += 1
    if ones < 40 or zeroes < 40: raise Exception("Clock not ticking")

async def loopback(ctl, o_pin, i_pin):
    print(f"Loopback {o_pin.name} => {i_pin.name}")
    i_pin.output_enable(False)
    o_pin.output_enable(True)
    for v in (0, 1, 0):
        o_pin.set_value(v)
        await ctl.cycle() # drive
        await ctl.cycle() # sample
        if i_pin.get_value() != v:
            raise Exception(f"Loopback {o_pin.name} => {i_pin.name} failed")

async def main():
    drv = ebyst.drivers.MPSSE(ebyst.drivers.MPSSE.list_devices([(0x1514, 0x2008)])[0])
    dev = ebyst.Device.from_bsdl("bsdl/MPF300TSFCG1152.bsdl")
    ctl = ebyst.TapController(drv)

    ctl.detect_chain()
    ctl.add_device(dev)
    ctl.validate_chain()

    try:
        ctl.extest()
        async with asyncio.TaskGroup() as tg:
            tg.create_task(leds(ctl, dev))
            tg.create_task(flash(ctl, dev))
            tg.create_task(mdio(ctl, dev))
            tg.create_task(ddr3(ctl, dev))
            tg.create_task(ddr4(ctl, dev))
            tg.create_task(clock(ctl, dev))

        ctl.extest_pulse()
        async with asyncio.TaskGroup() as tg:
            tg.create_task(loopback(ctl, dev.pinmap['IO_AH33'], dev.pinmap['IO_AG31']))
            tg.create_task(loopback(ctl, dev.pinmap['IO_AF33'], dev.pinmap['IO_AH29']))
    except KeyboardInterrupt:
        pass
    finally:
        ctl.reset()

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    asyncio.run(main())


# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import time
from bitarray import bitarray
from bitarray.util import ba2int, int2ba

import pyftdi
from pyftdi.ftdi import Ftdi

from .driver import Driver

class MPSSE(Driver):
    def __init__(self, url):
        Driver.__init__(self)

        self.url = url
        self.ftdi = Ftdi()
        self.ftdi.open_mpsse_from_url(self.url, direction=1|2|8, initial=0, frequency=1e6, latency=1)
        self.ftdi.reset()

    def set_freq(self, freq):
        freq_orig = freq
        while self.ftdi.set_frequency(freq) > freq_orig:
            freq *= 0.9

    def _read_bytes(self, count):
        r = bytearray()
        while len(r) < count:
            r += self.ftdi.read_data(count - len(r))
        return r

    @staticmethod
    def list_devices(custom_product_ids=[]):
        for custom_product_id in custom_product_ids:
            Ftdi.add_custom_vendor(custom_product_id[0])
            Ftdi.add_custom_product(custom_product_id[0], custom_product_id[1])

        urls = []
        for dev, ifaces in Ftdi.list_devices():
            for iface in range(ifaces):
                if not dev.sn is None:
                    urls.append(f"ftdi://0x{dev.vid:04x}:0x{dev.pid:04x}:{dev.sn}/{iface+1}")
                else:
                    urls.append(f"ftdi://0x{dev.vid:04x}:0x{dev.pid:04x}/{iface+1}")
        return urls

    def transfer(self, tms, tdi):
        self.ftdi.write_data(bytearray((Ftdi.RW_BITS_TMS_PVE_NVE, 0, (0x80 if tdi else 0) | (1 if tms else 0))))
        rd = self._read_bytes(1)
        return (rd[0] & 0x80) >> 7

    def transmit_tms_str(self, tms_str: bitarray, tdi=0):
        tdi = 0x80 if tdi else 0
        w = bytearray()
        for i in range(0, len(tms_str), 7):
            part = tms_str[i:i+7].copy()
            w += bytearray((Ftdi.WRITE_BITS_TMS_NVE, len(part)-1, tdi | ba2int(part)))
        self.ftdi.write_data(w)

    def transmit_tdi_str(self, tdi_str: bitarray, first_tms=0, last_tms=None):
        if last_tms is None: last_tms = first_tms
        if len(tdi_str) < 1: raise ValueError("n must be > 0")
        if len(tdi_str) == 1 and first_tms != last_tms: raise ValueError("last_tms must be first_tms when n == 1")

        tdi_str = tdi_str.copy()
        last_tdi = tdi_str.pop()
        w = bytearray()
        if len(tdi_str) > 0:
            w += bytearray((Ftdi.WRITE_BITS_TMS_NVE, 0, (tdi_str[0] << 7) | (1 if first_tms else 0)))
            for i in range(1, len(tdi_str), 8):
                part = tdi_str[i:i+8].copy()
                w += bytearray((Ftdi.WRITE_BITS_NVE_LSB, len(part)-1, ba2int(part)))
        w += bytearray((Ftdi.WRITE_BITS_TMS_NVE, 0, (last_tdi << 7) | (1 if last_tms else 0)))
        self.ftdi.write_data(w)

    def transfer_tdi_tdo_str(self, tdi_str: bitarray, first_tms=0, last_tms=0) -> bitarray:
        if last_tms is None: last_tms = first_tms
        if len(tdi_str) < 1: raise ValueError("n must be > 0")
        if len(tdi_str) == 1 and first_tms != last_tms: raise ValueError("last_tms must be first_tms when n == 1")

        tdi_str = tdi_str.copy()
        last_tdi = tdi_str.pop()

        w = bytearray()
        # NOTE: write all commands first, read results below for highest throughput,
        #       requires enough buffer size on FTDI/PC, not sure if this is an issue

        i = 0
        if len(tdi_str) > 0:
            w += bytearray((Ftdi.RW_BITS_TMS_PVE_NVE, 0, (0x80 if tdi_str[0] else 0) | (1 if first_tms else 0)))
            i += 1
        for i in range(1, len(tdi_str), 8):
            part = tdi_str[i:i+8].copy()
            w += bytearray((Ftdi.RW_BITS_PVE_NVE_LSB, len(part)-1, ba2int(part)))
        w += bytearray((Ftdi.RW_BITS_TMS_PVE_NVE, 0, (0x80 if last_tdi else 0) | (1 if last_tms else 0)))
        self.ftdi.write_data(w)

        r = bitarray(endian='little')
        if len(tdi_str) > 0:
            r.append((self._read_bytes(1)[0] & 0x80) >> 7)
        for i in range(1, len(tdi_str), 8):
            part = tdi_str[i:i+8]
            r += int2ba(self._read_bytes(1)[0] >> (8 - len(part)), len(part), 'little')
        r.append((self._read_bytes(1)[0] & 0x80) >> 7)
        return r

    def receive_tdo_str(self, n, first_tms=0, first_tdi=0, last_tms=None, last_tdi=None) -> bitarray:
        r = bitarray(endian='little')
        if last_tms is None: last_tms = first_tms
        if last_tdi is None: last_tdi = first_tdi
        if n < 1: raise ValueError("n must be > 0")
        if n == 1 and first_tms != last_tms: raise ValueError("last_tms must be first_tms when n == 1")
        if n == 1 and first_tdi != last_tdi: raise ValueError("last_tdi must be first_tdi when n == 1")
        if n > 1:
            r.append(self.transfer(first_tms, first_tdi))
            n = n - 1
        while n > 256+8:
            self.ftdi.write_data(bytearray((Ftdi.READ_BYTES_PVE_LSB, 31, 0)))
            for x in self._read_bytes(32):
                r += int2ba(x, 8, 'little')
            n = n - 256
        while n > 8:
            self.ftdi.write_data(bytearray((Ftdi.READ_BYTES_PVE_LSB, 0, 0)))
            for x in self._read_bytes(1):
                r += int2ba(x, 8, 'little')
            n = n - 8
        if n > 1:
            self.ftdi.write_data(bytearray((Ftdi.READ_BITS_PVE_LSB, n - 2)))
            x = self._read_bytes(1)
            r += int2ba(x[0] >> (9 - n), n - 1, 'little') # TODO; is this right?
            n = 1
        if n > 0:
            r.append(self.transfer(last_tms, last_tdi))
            n = n - 1
        return r

    def __repr__(self):
        return self.url

    @staticmethod
    def scan():
        for dev, interfaces in Ftdi.list_devices():
            for i in range(interfaces):
                yield f"ftdi://{dev.sn}/{i+1}: {dev.description}"


    def test(self):
        import random
        self.ftdi.write_data(bytearray((Ftdi.LOOPBACK_START, )))
        try:
            for i in range(100):
                tdi = random.randint(0, 1)
                tms = random.randint(0, 1)
                tdo = self.transfer(tms, tdi)
                assert tdo == tdi

            self.transmit_tms_str(bitarray("11001"), 0)
            self.transmit_tms_str(bitarray("1111111100000001"))

            self.transmit_tdi_str(bitarray("11001"))
            self.transmit_tdi_str(bitarray("1111111100000001"))

            for i in ("0", "1", "01", "10", "001", "100", "110011001", "00110011010011"):
                i = bitarray(i)
                o = self.transfer_tdi_tdo_str(bitarray(i))
                assert i == o
            for i in range(50):
                tdi = bitarray()
                for _ in range(random.randint(1, 200)):
                    tdi.append(random.randint(0, 1))
                tdo = self.transfer_tdi_tdo_str(tdi)
                assert tdi == tdo

            for tms in (0, 1):
                for first_tdi in (0, 1):
                    for last_tdi in (0, 1):
                        for n in (1, 2, 3, 4, 7, 8, 9, 10, 20, 30, 80, 1000):
                            if n == 1: last_tdi = first_tdi
                            tdo = self.receive_tdo_str(n, first_tms=tms, first_tdi=first_tdi, last_tdi=last_tdi)
                            assert len(tdo) == n
                            assert tdo[0] == first_tdi
                            assert tdo[-1] == last_tdi

        finally:
            self.ftdi.write_data(bytearray((Ftdi.LOOPBACK_END, )))



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from bitarray import bitarray

from ..device import Pin
from .spi import SPI

class MT25Q:
    def __init__(self, ctl, C: Pin, Sn: Pin, DQ0: Pin, DQ1: Pin, RESETn: Pin=None, WPn: Pin=None, HOLDn: Pin=None):
        self.ctl = ctl
        self.WPn = WPn
        self.RESETn = RESETn
        self.HOLDn = HOLDn
        self.spi = SPI(ctl, SCK=C, SSn=Sn, MOSI=DQ0, MISO=DQ1)

    async def init(self):
        if not self.RESETn is None:
            self.RESETn.output_enable(True)
            self.RESETn.set_value(0)
        if not self.WPn is None:
            self.WPn.output_enable(True)
            self.WPn.set_value(1)
        if not self.HOLDn is None:
            self.HOLDn.output_enable(True)
            self.HOLDn.set_value(1)
        await self.ctl.cycle()
        await self.spi.init()
        await self.ctl.cycle()
        if not self.RESETn is None:
            self.RESETn.set_value(1)
            await self.ctl.cycle()

    async def read_id(self):
        cmd = bitarray("10011110" + 20 * "00000000")
        data = await self.spi.transfer(cmd)
        return data[8:].tobytes()



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import re
from pprint import pprint

from bitarray import bitarray

from .bsdl import BSDLFile

SPACE = "[ \r\n\t]*"

RE_OPCODE = re.compile(f"{SPACE}(?P<instruction>[A-Za-z][A-Za-z_0-9]*){SPACE}\\((?P<opcode>[01]+(,{SPACE}[01]+)*)\\){SPACE}(,)?")
RE_CELL = re.compile(f"{SPACE}(?P<index>[0-9]+){SPACE}\\((?P<config>([^\\)\\(]*(\\([^\\)]*\\))*)*)\\)({SPACE},)?")

logger = logging.getLogger(__name__)

class StdLogicPattern:
    """Bit pattern supporting std_logic values"""
    def __init__(self, pattern):
        for c in pattern.upper():
            if not c in "01X": raise Exception(f"{c} not supported in bit pattern")
        self.pattern = pattern.upper()

    def __eq__(self, other):
        if len(self.pattern) != len(other): return False
        for c1, c2 in zip(self.pattern, other):
            if c1 != "X" and int(c1) != int(c2): return False
        return True

    def __str__(self):
        return f"StdLogicPattern('{self.pattern}')"

    def to_bitarray(self):
        return bitarray(self.pattern.replace("X", "0"), endian='little')

class Cell:
    """Represents a boundary scan cell"""
    def __init__(self, num, cell, port, function, safe, ctl_cell=None, out_dis_ctl=None, out_dis_val=None):
        self.num = num
        self.cell = cell
        self.port = port
        self.function = function
        self.safe = safe
        self.ctl_cell = int(ctl_cell) if not ctl_cell is None else None
        self.out_dis_ctl = int(out_dis_ctl) if not out_dis_ctl is None else None
        self.out_dis_val = out_dis_val
        self.in_value = None
        self.out_value = 0

        self.set_safe()

    def set_safe(self):
        if self.safe.upper() != 'X':
            self.out_value = int(self.safe)

    def __repr__(self):
        return f"{self.cell} @ {self.num}"

    @classmethod
    def parse(cls, num, parameters):
        return cls(num, *[p.strip() for p in parameters.split(",")])

class Pin:
    """Represents a device pin"""
    def __init__(self, name):
        self.name = name
        self.input_cell = None
        self.output_cell = None
        self.control_cell = None

    def output_enabled(self):
        if self.output_cell is None:
            return False
        elif self.control_cell is None:
            if self.input_cell is None:
                return True
            else:
                raise Exception(f"Pin {self.name} has no control cell")
        else:
            return self.control_cell.out_value != self.output_cell.out_dis_ctl
    
    def output_enable(self, enable=True):
        if enable:
            if self.output_cell is None:
                raise Exception(f"Pin {self.name} has no output cell")
            elif self.control_cell is None:
                if not self.input_cell is None:
                    raise Exception(f"Pin {self.name} has no control cell")
            else:
                self.control_cell.out_value = [1, 0][self.output_cell.out_dis_ctl]
        else:
            if self.input_cell is None:
                raise Exception(f"Pin {self.name} has no input cell")
            elif self.control_cell is None:
                if not self.output_cell is None:
                    raise Exception(f"Pin {self.name} has no control cell")
            elif not self.output_cell is None:
                self.control_cell.out_value = self.output_cell.out_dis_ctl

    def set_value(self, value):
        if self.output_cell is None:
            raise Exception(f"Pin {self.name} has no output cell")
        self.output_cell.out_value = 1 if value else 0

    def get_value(self):
        if self.input_cell is None:
            raise Exception(f"Pin {self.name} has no input cell")
        return self.input_cell.in_value

    def __repr__(self):
        if self.output_enabled():
            return f"<PIN {self.name}: output: {self.output_cell.out_value}>"
        else:
            return f"<PIN {self.name}: input>: {self.input_cell.in_value}>"

class DiffPin(tuple):
    """Differential pin pair"""
    def __new__(cls, p: Pin, n: Pin):
        return tuple.__new__(cls, (p, n))

    @property
    def name(self):
        return self[0].name

    def output_enable(self, enable=True):
        self[0].output_enable(enable)
        self[1].output_enable(enable)

    def output_enabled(self):
        return self[0].output_enabled()

    def set_value(self, value):
        self[0].set_value(1 if value else 0)
        self[1].set_value(0 if value else 1)

    def get_value(self):
        return self[0].get_value()

    def __repr__(self):
        if self.output_enabled():
            return f"<DIFFPIN {self.name}: output: {self.get_value()}>"
        else:
            return f"<DIFFPIN {self.name}: input>: {self.get_value()}>"

class PinGroup(list):
    """Group of pins to be read/written all at once"""
    def __init__(self, initial):
        list.__init__(self, initial)

    @property
    def name(self):
        return self[0].name

    def output_enable(self, enable=True):
        for pin in self:
            pin.output_enable(enable)

    def output_enabled(self):
        return self[0].output_enabled()

    def set_value(self, value):
        for (i, pin) in enumerate(self):
            try:
                pin.set_value(value[i])
            except TypeError:
                pin.set_value((value >> i) & 1)

    def get_value(self):
        r = bitarray(endian='little')
        for pin in self:
            r.append(pin.get_value())
        return r

class Device:
    def __init__(self, irlen, max_freq=None, idcode=None, opcodes=None, cells=[]):
        self.irlen = irlen
        self.max_freq = max_freq
        self.idcode = idcode
        if opcodes is None: opcodes = {'BYPASS': bitarray('1' * irlen, endian='little')}
        if not 'BYPASS' in opcodes: raise ValueError("BYPASS command is required")
        self.opcodes = opcodes
        self.cells = cells
        self.pinmap = {}
        for cell in self.cells:
            if cell.port != "*":
                try:
                    pin = self.pinmap[cell.port]
                except KeyError:
                    pin = Pin(cell.port)
                    self.pinmap[cell.port] = pin

                if cell.function in ("output2", "output3", "bidir"):
                    pin.output_cell = cell
                    if not cell.ctl_cell is None:
                        pin.control_cell = self.cells[cell.ctl_cell]
                if cell.function in ("input", "observe_only", "bidir", "clock"):
                    pin.input_cell = cell

    def reset(self):
        for cell in self.cells:
            cell.set_safe()

    def update_br(self, br):
        if len(br) != len(self.cells): raise ValueError("Invalid br length")
        for i, v in enumerate(br):
            self.cells[i].in_value = v

    def generate_br(self):
        r = bitarray(endian='little')
        for cell in self.cells:
            r.append(cell.out_value)
        return r

    @staticmethod
    def from_bsdl(fn):
        with open(fn, "rt") as f:
            bsdi_file = BSDLFile.parse(f)

        max_freq = float(bsdi_file.attributes["TAP_SCAN_CLOCK"].value)

        irlen = int(bsdi_file.attributes['INSTRUCTION_LENGTH'].value)
        idcode = StdLogicPattern(bsdi_file.attributes['IDCODE_REGISTER'].value[::-1])

        opcodes = {}
        opcode_str = bsdi_file.attributes['INSTRUCTION_OPCODE'].value
        while True:
            m = RE_OPCODE.match(opcode_str)
            if m:
                opcode = m['opcode'].split(",")
                ba = bitarray(opcode[0].strip(), endian='little')
                ba.reverse()
                opcodes[m['instruction'].upper()] = ba
                opcode_str = opcode_str[m.end():]
            else:
                break
        if len(opcode_str) != 0: raise Exception("Invalid INSTRUCTION_OPCODE format")

        brlen = int(bsdi_file.attributes['BOUNDARY_LENGTH'].value)

        cells = [None] * brlen
        cell_str = bsdi_file.attributes['BOUNDARY_REGISTER'].value.strip()
        while True:
            m = RE_CELL.match(cell_str)
            if m:
                cell = Cell.parse(int(m['index']), m['config'])
                cells[cell.num] = cell
                cell_str = cell_str[m.end():]
            else:
                break
        if len(cell_str) != 0: raise Exception("Invalid BOUNDARY_REGISTER format")

        return Device(irlen=irlen, max_freq=max_freq, idcode=idcode, opcodes=opcodes, cells=cells)



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import pyparsing as pp

logger = logging.getLogger(__name__)

class BSDLFile:
    """BSDL parser"""

    class Declaration:
        def __init__(self, name, type_, value=None, direction=None, range_start=None, range_end=None, owner=None):
            self.name = name
            self.type_ = type_
            self.value = value
            self.direction = direction
            self.range_start = range_start
            self.range_end = range_end
            self.owner = owner

    def __init__(self, name):
        self.name = name
        self.generics = {}
        self.ports = {}
        self.constants = {}
        self.attributes = {}

    @classmethod
    def parse(cls, f):
        comments = "--" + pp.SkipTo(pp.LineEnd())
        identifier = pp.Word(init_chars=pp.srange("[a-zA-Z]"), body_chars=pp.srange("[a-zA-Z0-9_]"))
        string_literal = pp.Combine(pp.QuotedString("\"") +
                                    pp.ZeroOrMore(pp.Suppress(pp.Literal("&")) + pp.QuotedString("\"")), adjacent=False)
        numeric_literal = pp.Or((pp.pyparsing_common.sci_real, pp.pyparsing_common.integer))
        primary = pp.Forward()
        enumeration_literal = (pp.Suppress(pp.Literal("(")) + primary +
                               pp.ZeroOrMore(pp.Suppress(pp.Literal(",")) + primary) + pp.Suppress(pp.Literal(")")))
        literal = pp.Or((numeric_literal, enumeration_literal, string_literal))
        primary <<= pp.Or((identifier, literal))
        expression = primary
        mode = pp.Or((pp.CaselessKeyword("in"), pp.CaselessKeyword("out"), pp.CaselessKeyword("inout"),
                     pp.CaselessKeyword("buffer"), pp.CaselessKeyword("linkage")))
        range = (pp.Suppress(pp.Literal("(")) + pp.pyparsing_common.integer +
                 pp.Or((pp.CaselessKeyword("to"), pp.CaselessKeyword("downto"))) +
                 pp.pyparsing_common.integer + pp.Suppress(pp.Literal(")")))
        subtype_indication = identifier + pp.Optional(range)
        declaration = pp.Group(identifier + pp.Suppress(pp.Literal(":")) +
                            pp.Optional(mode) + subtype_indication +
                            pp.Optional(pp.Suppress(pp.Literal(":=")) + expression))
        interface_list = declaration + pp.ZeroOrMore(pp.Suppress(pp.Literal(";")) + declaration)
        generic_clause = pp.Group(pp.CaselessKeyword("generic") + pp.Suppress(pp.Literal("(")) + interface_list +
                                  pp.Suppress(pp.Literal(")")) + pp.Suppress(pp.Literal(";")))
        port_clause = pp.Group(pp.CaselessKeyword("port") + pp.Suppress(pp.Literal("(")) + interface_list +
                               pp.Suppress(pp.Literal(")")) + pp.Suppress(pp.Literal(";")))
        use_clause = pp.Group(pp.CaselessKeyword("use") + identifier + pp.Literal(".") + identifier +
                              pp.ZeroOrMore(pp.Literal(",") + identifier) +
                              pp.Suppress(pp.Literal(";")))
        class_ = pp.Or((pp.CaselessKeyword("entity"), pp.CaselessKeyword("signal")))
        attribute_specification = pp.Group(pp.CaselessKeyword("attribute") + identifier +
                                           pp.Suppress(pp.CaselessKeyword("of")) + identifier +
                                           pp.Suppress(pp.Literal(":")) + class_ +
                                           pp.Suppress(pp.CaselessKeyword("is")) + expression +
                                           pp.Suppress(pp.Literal(";")))
        constant_declaration = pp.Group(pp.CaselessKeyword("constant") + declaration + pp.Suppress(pp.Literal(";")))
        entity_header = pp.Optional(generic_clause) + pp.Optional(port_clause)
        entity_declarative_item = pp.Or((use_clause, attribute_specification, constant_declaration))
        entity_declaration = (pp.Suppress(pp.CaselessKeyword("entity")) + identifier + pp.Suppress(pp.CaselessKeyword("is")) +
                              entity_header + pp.ZeroOrMore(entity_declarative_item) +
                              pp.Suppress(pp.CaselessKeyword("end") + pp.Optional(pp.CaselessKeyword("entity")) +
                                          pp.Optional(identifier)) +
                              pp.Suppress(pp.Literal(";")))
        bsdl_file = entity_declaration + pp.StringEnd()

        bsdl_file.ignore(comments)

        parsed = bsdl_file.parse_string(f.read())
        # parsed.pprint()
        r = cls(parsed[0])
        for item in parsed[1:]:
            if item[0] == "generic":
                for generic in item[1:]:
                    assert len(generic) == 3
                    generic = BSDLFile.Declaration(generic[0], generic[1], value=generic[2])
                    r.generics[generic.name] = generic
            elif item[0] == "port":
                for port in item[1:]:
                    if len(port) == 3:
                        port = BSDLFile.Declaration(port[0], port[2], direction=port[1])
                    elif len(port) == 6:
                        port = BSDLFile.Declaration(port[0], port[2], direction=port[1],
                                                    range_start=port[3], range_end=port[5])
                    r.ports[port.name] = port
            elif item[0] == "use":
                pass
            elif item[0] == "constant":
                assert len(item) == 2
                assert len(item[1]) == 3
                constant = BSDLFile.Declaration(name=item[1][0], type_=item[1][1], value=item[1][2])
                r.constants[constant.name] = constant
            elif item[0] == "attribute":
                assert len(item) >= 5
                # TODO fix tuples
                attribute = BSDLFile.Declaration(name=item[1], type_ = item[3], owner=item[2], value=item[4])
                r.attributes[attribute.name] = attribute
            else:
                assert False

        return r



# Copyright (c) 2025 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from binascii import hexlify
from bitarray import bitarray

from . import errors

class VariableScope(dict):
    def __getitem__(self, key: str):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise errors.VariableNotDefined(key) from None

class Evaluatable:
    def evaluate(self, scope=VariableScope()):
        raise NotImplementedError()

    def optimize(self):
        return self

class Array(Evaluatable):
    def __getitem__(self, k):
        raise NotImplementedError()

    def __setitem__(self, k, v):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

class Variable(Evaluatable):
    def __init__(self, v):
        assert isinstance(v, Evaluatable)
        self.v = v

    def assign(self, v, *args, **kwargs):
        if not isinstance(self.v, Any):
            v = type(self.v)(v) # type: ignore
        assert isinstance(v, Evaluatable)
        self.v = v

    def evaluate(self, scope=VariableScope()):
        return self.v.evaluate(scope)

class ArrayVariable(Variable, Array):
    def __init__(self, v):
        assert isinstance(v, Array)
        len(v)
        self.v = v

    def assign(self, pos_or_v, v=None):
        if v is None:
            slice_ = None
            v = pos_or_v
        elif isinstance(pos_or_v, int) or isinstance(pos_or_v, slice):
            slice_ = pos_or_v
        else:
            assert False

        self.v.__setitem__(slice_, v)

    def evaluate(self, scope=VariableScope()):
        return self.v.evaluate(scope)

    def __len__(self):
        return len(self.v)

    def __str__(self):
        return str(self.v)

class CheckedVariableScope(VariableScope):
    def __setitem__(self, key: str, value: Variable) -> None:
        assert isinstance(value, Variable)
        return super().__setitem__(key, value)

class Int(Evaluatable):
    def __init__(self, v):
        if (isinstance(v, int) or isinstance(v, str)) and not isinstance(v, bool):
            self.v = int(v)
        elif isinstance(v, Any) or isinstance(v, Int):
            self.v = v.v
        else:
            raise errors.StaplValueError(f"Could not convert {repr(v)} to Int")

    def evaluate(self, scope=VariableScope()):
        return self

    def __add__(self, other):
        return Int(self.v + Int(other).v)

    def __sub__(self, other):
        return Int(self.v - Int(other).v)

    def __mul__(self, other):
        return Int(self.v * Int(other).v)

    def __floordiv__(self, other):
        return Int(self.v // Int(other).v)

    def __mod__(self, other):
        return Int(self.v % Int(other).v)

    def __lshift__(self, other):
        return Int(self.v << Int(other).v)

    def __rshift__(self, other):
        return Int(self.v >> Int(other).v)

    def __and__(self, other):
        return Int(self.v & Int(other).v)

    def __or__(self, other):
        return Int(self.v | Int(other).v)

    def __xor__(self, other):
        return Int(self.v ^ Int(other).v)

    def __ge__(self, other):
        return Bool(self.v >= Int(other).v)

    def __gt__(self, other):
        return Bool(self.v > Int(other).v)

    def __le__(self, other):
        return Bool(self.v <= Int(other).v)

    def __lt__(self, other):
        return Bool(self.v < Int(other).v)

    def __eq__(self, other): # type: ignore
        return Bool(self.v == Int(other).v)

    def __ne__(self, other): # type: ignore
        return Bool(self.v != Int(other).v)

    def __invert__(self):
        return Int(~self.v)

    def __neg__(self):
        return Int(-self.v)

    def __str__(self):
        return str(self.v)

    def __int__(self):
        return self.v

    def __repr__(self):
        return f"Int({self.v})"

    def clone(self):
        return Int(self)

class Bool(Evaluatable):
    def __init__(self, v):
        if (isinstance(v, int) or isinstance(v, str)) and int(v) in (0, 1):
            self.v = int(v)
        elif (isinstance(v, Any) or isinstance(v, Bool)) and v.v in (0, 1):
            self.v = v.v
        elif isinstance(v, bitarray) and len(v) == 1:
            self.v = 1 if v[0] else 0
        else:
            raise errors.StaplValueError(f"Could not convert {repr(v)} to Bool")

    def evaluate(self, scope=VariableScope()):
        return self

    def __eq__(self, other): # type: ignore
        return Bool(self.v == Bool(other).v)

    def __ne__(self, other): # type: ignore
        return Bool(self.v != Bool(other).v)

    def __and__(self, other):
        return Bool(self.v and Bool(other).v)

    def __or__(self, other):
        return Bool(self.v or Bool(other).v)

    def __xor__(self, other):
        return Bool(self.v ^ Bool(other).v)

    def __invert__(self):
        return Bool(0 if self.v else 1)

    def __bool__(self):
        return bool(self.v)

    def __str__(self):
        return str(self.v)

    def __repr__(self):
        return f"Bool({self.v})"

    def clone(self):
        return Bool(self.v)

class Any(Int):
    """Boolean or integer"""
    def __init__(self, v):
        if isinstance(v, int) or isinstance(v, str) or isinstance(v, Any):
            self.v = int(v)
        else:
            raise errors.StaplValueError(f"Could not convert {repr(v)} to Any")

    def evaluate(self, scope=VariableScope()):
        return self

    def __eq__(self, other):
        if isinstance(other, Bool):
            return Bool(self) == other
        else:
            return Int(self) == other

    def __ne__(self, other):
        if isinstance(other, Bool):
            return Bool(self) != other
        else:
            return Int(self) != other

    def __and__(self, other):
        return NotImplemented

    def __or__(self, other):
        return NotImplemented

    def __xor__(self, other):
        return NotImplemented

    def __str__(self):
        return str(self.v)

    def __repr__(self):
        return f"Any({self.v})"

    def __int__(self):
        return self.v

    def __bool__(self):
        return bool(self.v)

    def clone(self):
        return Any(self)

class IntArray(list, Array):
    def __init__(self, init=[]):
        for x in init:
            if isinstance(x, int):
                self.append(Int(x))
            elif isinstance(x, Int):
                self.append(x)
            else:
                raise errors.StaplValueError(f"Can't convert {x} to Int")

    def __getitem__(self, i): # type: ignore
        if isinstance(i, int):
            return Int(super().__getitem__(i))
        elif isinstance(i, slice):
            assert i.step is None

            if i.start <= i.stop:
                return IntArray(super().__getitem__(slice(i.start, i.stop+1)))
            else:
                return IntArray(super().__getitem__(slice(i.start, i.stop-1 if i.stop > 0 else None, -1)))
        else:
            raise TypeError(f"Invalid type {type(i)} for slice")

    def __setitem__(self, i, v):
        if i is None:
            if not isinstance(v, IntArray):
                raise errors.StaplValueError(f"Can't assign {type(v)} to integer array")
            super().__setitem__(slice(0, len(self)), v)
        elif isinstance(i, int):
            if not isinstance(v, Int):
                raise errors.StaplValueError(f"Can't assign {v} to integer array element")
            super().__setitem__(i, v)
        elif isinstance(i, slice):
            assert i.step is None
            if not isinstance(v, IntArray):
                TypeError(f"Can't assign {type(v)} to integer array")

            try:
                if i.start <= i.stop:
                    super().__setitem__(slice(i.start, i.stop+1), v)
                else:
                    super().__setitem__(slice(i.start, i.stop-1 if i.stop > 0 else None, -1), v)
            except (TypeError, ValueError) as e:
                raise errors.StaplValueError(str(e)) from None
        else:
            raise TypeError(f"Invalid type {type(i)} for slice")

    def evaluate(self, scope=VariableScope()):
        return self

class BoolArray(Array):
    def __init__(self, v):
        self.v = bitarray(v, endian='little')
        if isinstance(v, str):
            self.v.reverse()

    def __getitem__(self, i):
        if isinstance(i, int):
            return Bool(self.v[i])
        elif isinstance(i, slice):
            assert i.step is None

            if i.start < i.stop:
                x = self.v.__getitem__(slice(i.stop, i.start-1 if i.start > 0 else None, -1))
                return BoolArray(x)
            else:
                x = self.v.__getitem__(slice(i.stop, i.start+1))
                return BoolArray(x)
        else:
            raise TypeError(f"Invalid type {type(i)} for slice")

    def __setitem__(self, i, v):
        if i is None:
            self.v.setall(0)
            self.__setitem__(slice(min(len(v), len(self))-1, 0), v)
        elif isinstance(i, int):
            if isinstance(v, Bool) or isinstance(v, Any):
                self.v.__setitem__(i, v.v)
            else:
                raise errors.StaplValueError(f"Can't assign {v} to boolean array element")
        elif isinstance(i, slice):
            assert i.step is None
            if not isinstance(v, BoolArray):
                raise errors.StaplValueError(f"Can't assign {type(v)} to boolean array")

            slice_length = i.stop - i.start + 1 if i.stop > i.start else i.start - i.stop + 1
            if len(v) > slice_length:
                v = BoolArray(v.v[:slice_length])
            elif len(v) < slice_length:
                v = BoolArray(v.v + bitarray("0", endian='little') * (slice_length - len(v)))

            if i.start < i.stop:
                self.v.__setitem__(slice(i.stop, i.start-1 if i.start > 0 else None, -1), v.v)
            else:
                self.v.__setitem__(slice(i.stop, i.start+1), v.v)
        else:
            raise TypeError(f"Invalid type {type(i)} for slice")

    def __eq__(self, other):
        if not isinstance(other, BoolArray): return False
        return self.v == other.v

    def __len__(self):
        return len(self.v)

    def __str__(self):
        return hexlify(self.v.tobytes()[::-1]).decode("ascii")

    def __repr__(self):
        return repr(self.v)

    def evaluate(self, scope=VariableScope()):
        return self

    def reverse(self):
        x = BoolArray(self.v)
        x.v.reverse()
        return x

    def to_bitarray(self):
        return self.v

    def extend(self, length):
        while len(self.v) < length: self.v.append(0)

    def __and__(self, other):
        return self.v & other.v

class String:
    def __init__(self, v):
        self.v = str(v)

    def __str__(self):
        return self.v

    def __repr__(self):
        return f"String({self.v})"



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import pyparsing as pp

from .data import Int, IntArray
from .expressions import Expression
from ..tap_controller import State

logger = logging.getLogger(__name__)

class Note:
    def __init__(self,  _s, _loc, tokens):
        assert len(tokens) == 2
        self.key = tokens[0]
        self.text = tokens[1]

    def __str__(self):
        return f"{self.key}: {self.text}"

class Action:
    def __init__(self,  _s, _loc, tokens):
        self.name = tokens[0]
        self.text = tokens[1][0] if len(tokens[1]) > 0 else tokens[0]
        self.procedures = []
        for procedure in tokens[2:]:
            self.procedures.append((procedure[0], procedure.opt))

    def __repr__(self):
        return f"<Action {self.text}: {', '.join(p[0] + ("?" if p[1] == "optional" else "") +
                                                 ("*" if p[1] == "recommended" else "") for p in self.procedures)}>"

    def __str__(self):
        return f"ACTION {self.name}"

class VariableDecl:
    def __init__(self,  _s, _loc, tokens):
        self.name = tokens[0]
        if len(tokens) > 1:
            self.length = tokens[1].evaluate()
        else:
            self.length = None

    def __str__(self):
        if self.length is None:
            return f"{self.name}"
        else:
            return f"{self.name}[{self.length}]"

class Variable:
    def __init__(self,  s, loc, tokens):
        if len(tokens) == 1:
            self.name = tokens[0]
            self.first = None
            self.last = None
        elif len(tokens) == 2:
            self.name = tokens[0]
            self.first = tokens[1]
            self.last = None
        elif len(tokens) == 3:
            self.name = tokens[0]
            self.first = tokens[1]
            self.last = tokens[2]
        else:
            assert False

    def __str__(self):
        return self.name

class Instruction:
    def __init__(self,  s, loc, tokens):
        self.line = pp.lineno(loc, s)
        self.col = pp.col(loc, s)

class LabelledInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 2
        self.label = tokens[0][0] if len(tokens[0]) > 0 else None
        self.instruction = tokens[1]

    def execute(self, interpreter):
        return self.instruction.execute(interpreter)

    def __str__(self):
        return f"{self.label}: {self.instruction}"

class AssignmentInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        if len(tokens) == 2:
            self.variable = tokens[0]
            self.value = tokens[1]
        else:
            print(tokens)
            assert False

    def __str__(self):
        return f"{self.variable.name} = {self.value}"

class BooleanInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.name = tokens[0].name
        self.length = None if tokens[0].length is None else int(tokens[0].length.evaluate())
        if len(tokens) > 1:
            self.value = tokens[1]
        else:
            self.value = None

    def __str__(self):
        return f"BOOLEAN {self.name}"

class CallInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 1
        self.procedure = tokens[0]

    def __str__(self):
        return f"CALL {self.procedure}"

class ScanInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.length = tokens[0]
        self.data_array = tokens[1]
        i = 2
        if len(tokens) > i and tokens[i].upper() == "CAPTURE":
            self.capture_array = tokens[i+1]
            i += 2
        else:
            self.capture_array = None
        if len(tokens) > i and tokens[i].upper() == "COMPARE":
            self.compare_array = tokens[i+1]
            self.compare_mask_array = tokens[i+2]
            self.compare_result = tokens[i+3]
            i += 4
        else:
            self.compare_array = None
            self.mask_array = None
            self.compare_result = None
        assert i == len(tokens)

class DrScanInstruction(ScanInstruction):
    def __str__(self):
        return f"DRSCAN {self.length}"

class DrStopInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 1
        self.state = tokens[0].state

    def __str__(self):
        return f"IRSTOP {self.state}"

class ExitInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.exit_code = tokens[0]

    def __str__(self):
        return f"EXIT {self.exit_code}"

class ExportInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.key = tokens[0]
        self.parts = tokens[1:]

    def __repr__(self):
        return f"EXPORT {', '.join(str(s) for s in self.parts)}"

class FrequencyInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.frequency = tokens[0]

    def __str__(self):
        return f"FREQUENCY {self.frequency}"

class GotoInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.label = tokens[1]

    def __str__(self):
        return f"GOTO {self.label}"

class IfInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.condition = tokens[0]
        self.instruction = tokens[1]

    def __str__(self):
        return f"IF {self.condition} THEN {self.instruction}"

class IntegerInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.name = tokens[0].name
        self.length = None if tokens[0].length is None else int(tokens[0].length.evaluate())

        if not self.length is None:
            self.value = IntArray([Int(0)] * self.length)
            if len(tokens) > 1:
                assert self.length == len(tokens) - 1
                for i in range(self.length):
                    self.value[i] = tokens[i+1]
        else:
            self.value = Int(0)
            if len(tokens) > 1:
                assert 1 == len(tokens) - 1
                self.value = tokens[1]

    def __repr__(self):
        if self.length is None:
            return f"<Integer {self.name}>"
        else:
            return f"<Integer {self.name}[{self.length}]>"

    def __str__(self):
        return f"INTEGER {self.name}"

class IrScanInstruction(ScanInstruction):
    def __str__(self):
        return f"IRSCAN {self.length}"

class IrStopInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 1
        self.state = tokens[0].state

    def __str__(self):
        return f"IRSTOP {self.state}"

class PopInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        if len(tokens) == 1:
            self.variable = tokens[0]
        else:
            assert False

    def __str__(self):
        return f"POP {self.variable}"

class PrintInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.parts = tokens

    def __str__(self):
        return f"PRINT {', '.join(str(s) for s in self.parts)}"

class PushInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 1
        self.value = tokens[0]

    def __str__(self):
        return f"PUSH {self.value}"

class StateConvert:
    def __init__(self,  s, loc, tokens):
        assert len(tokens) == 1
        if tokens[0] == "RESET":
            self.state = State.TEST_LOGIC_RESET
        elif tokens[0] == "IDLE":
            self.state = State.RUN_TEST_IDLE
        elif tokens[0] == "DRSELECT":
            self.state = State.SELECT_DR_SCAN
        elif tokens[0] == "DRCAPTURE":
            self.state = State.CAPTURE_DR
        elif tokens[0] == "DRSHIFt":
            self.state = State.SHIFT_DR
        elif tokens[0] == "DREXIT1":
            self.state = State.EXIT1_DR
        elif tokens[0] == "DRPAUSE":
            self.state = State.PAUSE_DR
        elif tokens[0] == "DREXIT2":
            self.state = State.EXIT2_DR
        elif tokens[0] == "DRUPDATE":
            self.state = State.UPDATE_DR
        elif tokens[0] == "IRSELECT":
            self.state = State.SELECT_IR_SCAN
        elif tokens[0] == "IRCAPTURE":
            self.state = State.CAPTURE_IR
        elif tokens[0] == "IRSHIFT":
            self.state = State.SHIFT_IR
        elif tokens[0] == "IREXIT1":
            self.state = State.EXIT1_IR
        elif tokens[0] == "IRPAUSE":
            self.state = State.PAUSE_IR
        elif tokens[0] == "IREXIT2":
            self.state = State.EXIT2_IR
        elif tokens[0] == "IRUPDATE":
            self.state = State.UPDATE_IR
        else:
            assert False

class StateInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        self.states = [token.state for token in tokens]

    def __str__(self):
        return f"STATE {', '.join(str(s) for s in self.states)}"

class WaitType:
    def __init__(self,  s, loc, tokens):
        self.usec = self.cycles = Int(0)
        if len(tokens) == 2 and tokens[1].upper() == "USEC":
            self.usec = tokens[0]
        elif len(tokens) == 2 and tokens[1].upper() == "CYCLES":
            self.cycles = tokens[0]
        elif len(tokens) == 4 and tokens[1].upper() == "CYCLES" and tokens[3].upper() == "USEC":
            self.cycles = tokens[0]
            self.usec = tokens[2]
        else:
            assert False

    def __str__(self):
        return f"{self.cycles} CYCLES {self.usec} USEC"

class TRSTInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        if len(tokens) == 0:
            self.wait_cycles = self.wait_usec = Int(0)
        elif len(tokens) == 1:
            self.wait_cycles = tokens[0].cycles
            self.wait_usec = tokens[0].usec
        else:
            assert False

    def __str__(self):
        return f"TRST {self.wait_cycles} CYCLES {self.wait_usec} USEC"

class WaitInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)

        self.wait_state = self.end_state = None
        self.wait_usec = self.wait_cycles = Int(0)

        i = 0
        if len(tokens) > i and not isinstance(tokens[i], WaitType):
            self.wait_state = tokens[i].state
            i += 1

        if len(tokens) > i and isinstance(tokens[i], WaitType):
            self.wait_cycles = tokens[i].cycles
            self.wait_usec = tokens[i].usec
            i += 1
        else:
            assert False

        if len(tokens) > i and not isinstance(tokens[i], WaitType):
            self.end_state = tokens[i].state
            i += 1

        if len(tokens) > i and isinstance(tokens[i], WaitType):
            raise Exception("MAX wait times are not supported")

        assert i == len(tokens)

    def __str__(self):
        r = "WAIT"
        if not self.wait_state is None:
            r += f" {self.wait_state}"
        r += f" {self.wait_cycles} CYCLES {self.wait_usec} USEC"
        if not self.wait_state is None:
            r += f" {self.end_state}"
        return r

class ForInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 3 or len(tokens) == 4
        self.var = tokens[0]
        self.start = tokens[1]
        self.end = tokens[2]
        self.step = tokens[3] if len(tokens) == 4 else Int(1)

    def __str__(self):
        return f"FOR {self.var} = {self.start}..{self.end} STEP {self.step}"

class NextInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 1
        self.var = tokens[0]

    def __str__(self):
        return f"NEXT {self.var}"

class ScanModifierInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        if len(tokens) == 1:
            self.bits = tokens[0]
            self.value = None
        elif len(tokens) == 2:
            self.bits = tokens[0]
            self.value = tokens[1]
        else:
            assert False

class PostDrInstruction(ScanModifierInstruction):
    pass

class PostIrInstruction(ScanModifierInstruction):
    pass

class PreDrInstruction(ScanModifierInstruction):
    pass

class PreIrInstruction(ScanModifierInstruction):
    pass

class ProcedureInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)
        assert len(tokens) == 2
        self.name = tokens[0]
        self.uses = tokens[1]

    def __repr__(self):
        return f"<Procedure {self.name}>"

    def __str__(self):
        return f"PROCEDURE {self.name}"

class EndProcedureInstruction(Instruction):
    def __init__(self,  s, loc, tokens):
        Instruction.__init__(self, s, loc, tokens)

    def __str__(self):
        return f"ENDPROC"

class DataInstruction(Instruction):
    def __init__(self, s, loc, tokens):
        assert len(tokens) == 1
        Instruction.__init__(self, s, loc, tokens)
        self.name = tokens[0]

    def __str__(self):
        return f"DATA {self.name}"

class EndDataInstruction(Instruction):
    def __str__(self):
        return f"<ENDDATA>"

class Crc:
    def __init__(self,  s, loc, tokens):
        assert len(tokens) == 1
        self.expected = int(tokens[0], 16)

        if self.expected != 0:
            CCITT_CRC =  0x8408
            crc_register = 0xFFFF
            for in_byte in s[:loc]:
                in_byte = ord(in_byte)
                if in_byte != 13:
                    for _ in range(8):
                        feedback = (in_byte ^ crc_register) & 0x01;
                        crc_register >>= 1; # shift the shift register
                        if feedback: crc_register ^= CCITT_CRC; # invert selected bits
                        in_byte >>= 1; # get the next bit of in_byte

            self.actual = (~crc_register) & 0xFFFF
        else:
            self.actual = 0

    def is_correct(self):
        return self.actual == self.expected

    def __repr__(self):
        if self.is_correct():
            return "<Crc (correct)>"
        else:
            return "<Crc (incorrect)>"

    def __str__(self):
        return f"CRC {self.expected:04x}"

class StaplFile:
    """STAPL parser"""

    def __init__(self,  tokens):
        self.actions = {}
        self.procedures = {}
        self.data_blocks = {}
        self.labels = {}
        self.notes = []
        self.statements = []
        procedure = data_block = None
        for token in tokens:
            if isinstance(token, Note):
                assert data_block is None and procedure is None
                logger.info(f"NOTE: {token}")
                self.notes.append(token)
            elif isinstance(token, Action):
                assert data_block is None and procedure is None
                logger.debug(f"Action: {token.name}")
                self.actions[token.name] = token
            elif isinstance(token, LabelledInstruction):
                if isinstance(token.instruction, ProcedureInstruction):
                    assert data_block is None
                    procedure = token.instruction.name
                    self.procedures[procedure] = len(self.statements)
                    self.labels[procedure] = {}
                elif isinstance(token.instruction, EndProcedureInstruction):
                    procedure = None
                elif isinstance(token.instruction, DataInstruction):
                    assert procedure is None
                    data_block = token.instruction.name
                    self.data_blocks[data_block] = len(self.statements)
                elif isinstance(token.instruction, EndDataInstruction):
                    data_block = None
                if not token.label is None:
                    assert not procedure is None
                    self.labels[procedure][token.label] = len(self.statements)
                self.statements.append(token)
            elif isinstance(token, Crc):
                if not token.is_correct():
                    logger.warning(f"CRC check failed ({token.actual:04x}, expected: {token.expected:04x})")
            else:
                assert False

    @classmethod
    def parse(cls, f):
        comments = "`" + pp.SkipTo(pp.LineEnd())
        instruction = pp.Forward()

        expression = Expression.get_parse_rule()
        identifier = pp.Word(init_chars=pp.srange("[a-zA-Z]"), body_chars=pp.srange("[a-zA-Z0-9_]"))
        variable_decl = (identifier + pp.Opt(pp.Literal("[").suppress() - expression - pp.Literal("]").suppress())).set_parse_action(VariableDecl)
        state_name = pp.MatchFirst((pp.CaselessKeyword("RESET"),
                                    pp.CaselessKeyword("IDLE"),
                                    pp.CaselessKeyword("DRSELECT"),
                                    pp.CaselessKeyword("DRCAPTURE"),
                                    pp.CaselessKeyword("DRSHIFT"),
                                    pp.CaselessKeyword("DREXIT1"),
                                    pp.CaselessKeyword("DRPAUSE"),
                                    pp.CaselessKeyword("DREXIT2"),
                                    pp.CaselessKeyword("DRUPDATE"),
                                    pp.CaselessKeyword("IRSELECT"),
                                    pp.CaselessKeyword("IRCAPTURE"),
                                    pp.CaselessKeyword("IRSHIFT"),
                                    pp.CaselessKeyword("IREXIT1"),
                                    pp.CaselessKeyword("IRPAUSE"),
                                    pp.CaselessKeyword("IREXIT2"),
                                    pp.CaselessKeyword("IRUPDATE"))).set_parse_action(StateConvert)

        str_expression = pp.Or((pp.QuotedString("\""), expression)) # TODO

        action = (pp.CaselessKeyword("ACTION").suppress() - identifier - pp.Group(pp.Opt(pp.QuotedString("\""))) -
                  pp.Literal("=").suppress() -
                  pp.Group(identifier - pp.Or((pp.CaselessKeyword("OPTIONAL") - pp.Tag("opt", "optional"),
                                               pp.CaselessKeyword("RECOMMENDED") - pp.Tag("opt", "recommended"),
                                               pp.Tag("opt", "required")))) -
                  pp.ZeroOrMore(pp.Literal(",").suppress() - pp.Group(identifier -
                                pp.Or((pp.CaselessKeyword("OPTIONAL") - pp.Tag("opt", "optional"),
                                       pp.CaselessKeyword("RECOMMENDED") - pp.Tag("opt", "recommended"),
                                       pp.Tag("opt", "required"))))) -
                  pp.Literal(";").suppress()).set_parse_action(Action)
        variable = (pp.Word(init_chars=pp.srange("[a-zA-Z]"), body_chars=pp.srange("[a-zA-Z0-9_]")) -
                    pp.Opt(pp.Literal("[").suppress() - pp.Opt(expression - pp.Opt(pp.Literal("..").suppress() - expression)) -
                           pp.Literal("]").suppress())).set_parse_action(Variable)
        assignment = (variable + pp.Literal("=").suppress() - expression - pp.Suppress(pp.Literal(";"))).set_parse_action(AssignmentInstruction)
        boolean = (pp.CaselessKeyword("BOOLEAN").suppress() - variable_decl - pp.Opt(pp.Literal("=").suppress() -
                           expression) - pp.Literal(";").suppress()).set_parse_action(BooleanInstruction)
        call = (pp.CaselessKeyword("CALL").suppress() - identifier - pp.Suppress(pp.Literal(";"))).set_parse_action(CallInstruction)
        crc = (pp.CaselessKeyword("CRC").suppress() - pp.Word(pp.srange("[0-9a-fA-F]")) - pp.Literal(";").suppress()).set_parse_action(Crc)
        data = (pp.CaselessKeyword("DATA").suppress() - identifier - pp.Suppress(pp.Literal(";"))).set_parse_action(DataInstruction)
        drscan = (pp.CaselessKeyword("DRSCAN").suppress() - expression - pp.Literal(",").suppress() - expression -
                          pp.Opt(pp.Literal(",").suppress() + pp.CaselessKeyword("CAPTURE") - variable) -
                          pp.Opt(pp.Literal(",").suppress() - pp.CaselessKeyword("COMPARE") - expression -
                                 pp.Literal(",").suppress() - expression + pp.Literal(",").suppress() - variable) -
                          pp.Literal(";").suppress()).set_parse_action(DrScanInstruction)
        drstop = (pp.CaselessKeyword("DRSTOP").suppress() - state_name - pp.Suppress(pp.Literal(";"))).set_parse_action(DrStopInstruction)
        end_data = (pp.CaselessKeyword("ENDDATA").suppress() - pp.Literal(";").suppress()).set_parse_action(EndDataInstruction)
        end_procedure = (pp.CaselessKeyword("ENDPROC").suppress() - pp.Literal(";").suppress()).set_parse_action(EndProcedureInstruction)
        exit = (pp.CaselessKeyword("EXIT").suppress() - expression - pp.Suppress(pp.Literal(";"))).set_parse_action(ExitInstruction)
        export = (pp.CaselessKeyword("EXPORT").suppress() - str_expression -
                  pp.ZeroOrMore(pp.Suppress(pp.Literal(",")) - str_expression) - pp.Literal(";").suppress()).set_parse_action(ExportInstruction)
        for_ = (pp.CaselessKeyword("FOR").suppress() - identifier - pp.Literal("=").suppress() - expression -
                pp.CaselessKeyword("TO").suppress() - expression -
                pp.Opt(pp.CaselessKeyword("STEP").suppress() - expression) - pp.Suppress(pp.Literal(";"))).set_parse_action(ForInstruction)
        frequency = (pp.CaselessKeyword("FREQUENCY").suppress() - pp.Opt(expression) - pp.Literal(";").suppress()).set_parse_action(FrequencyInstruction)
        goto = (pp.CaselessKeyword("GOTO") - identifier - pp.Suppress(pp.Literal(";"))).set_parse_action(GotoInstruction)
        if_ = (pp.CaselessKeyword("IF").suppress() - expression - pp.CaselessKeyword("THEN").suppress() - instruction).set_parse_action(IfInstruction)
        integer = (pp.CaselessKeyword("INTEGER").suppress() - variable_decl -
                           pp.Opt(pp.Literal("=").suppress() - expression - pp.ZeroOrMore(pp.Literal(",").suppress() - expression)) - pp.Literal(";").suppress()).set_parse_action(IntegerInstruction)
        irscan = (pp.CaselessKeyword("IRSCAN").suppress() - expression - pp.Literal(",").suppress() - expression -
                          pp.Opt(pp.Literal(",").suppress() + pp.CaselessKeyword("CAPTURE") - variable) -
                          pp.Opt(pp.Literal(",").suppress() - pp.CaselessKeyword("COMPARE") - expression -
                                 pp.Literal(",").suppress() - expression + pp.Literal(",").suppress() - variable) -
                          pp.Literal(";").suppress()).set_parse_action(IrScanInstruction)
        irstop = (pp.CaselessKeyword("IRSTOP").suppress() - state_name - pp.Suppress(pp.Literal(";"))).set_parse_action(IrStopInstruction)
        next =  (pp.CaselessKeyword("NEXT").suppress() - identifier - pp.Literal(";").suppress()).set_parse_action(NextInstruction)
        note = (pp.CaselessKeyword("NOTE").suppress() - pp.QuotedString("\"") - pp.QuotedString("\"") - pp.Literal(";").suppress()).set_parse_action(Note)
        pop = ((pp.CaselessKeyword("POP").suppress() - variable - pp.Suppress(pp.Literal(";")))).set_parse_action(PopInstruction)
        postdr = ((pp.CaselessKeyword("POSTDR").suppress() - expression - pp.Opt(pp.Literal(",") - expression) - pp.Literal(";").suppress())).set_parse_action(PostDrInstruction)
        postir = ((pp.CaselessKeyword("POSTIR").suppress() - expression - pp.Opt(pp.Literal(",") - expression) - pp.Literal(";").suppress())).set_parse_action(PostIrInstruction)
        predr = ((pp.CaselessKeyword("PREDR").suppress() - expression - pp.Opt(pp.Literal(",") - expression) - pp.Literal(";").suppress())).set_parse_action(PreDrInstruction)
        preir = ((pp.CaselessKeyword("PREIR").suppress() - expression - pp.Opt(pp.Literal(",") - expression) - pp.Literal(";").suppress())).set_parse_action(PreIrInstruction)
        print_ = (pp.CaselessKeyword("PRINT").suppress() - str_expression -
                  pp.ZeroOrMore(pp.Suppress(pp.Literal(",")) - str_expression) - pp.Literal(";").suppress()).set_parse_action(PrintInstruction)
        procedure = (pp.CaselessKeyword("PROCEDURE").suppress() - identifier -
                     pp.Group(pp.Opt(pp.CaselessKeyword("USES").suppress() - identifier - pp.ZeroOrMore(pp.Literal(",").suppress() - identifier))) -
                     pp.Literal(";").suppress()).set_parse_action(ProcedureInstruction)
        push = (pp.CaselessKeyword("PUSH").suppress() - expression - pp.Suppress(pp.Literal(";"))).set_parse_action(PushInstruction)
        state = (pp.CaselessKeyword("STATE").suppress() - pp.OneOrMore(state_name) - pp.Literal(";").suppress()).set_parse_action(StateInstruction)
        wait_type = pp.Or((expression - pp.CaselessKeyword("CYCLES") -
                           pp.Opt(pp.Suppress(pp.Literal(",")) + expression + pp.CaselessKeyword("USEC")),
                           expression - pp.CaselessKeyword("USEC"))).set_parse_action(WaitType)
        trst = (pp.CaselessKeyword("TRST").suppress() - pp.Opt(wait_type) - pp.Suppress(pp.Literal(";"))).set_parse_action(TRSTInstruction)
        wait = (pp.CaselessKeyword("WAIT").suppress() - pp.Opt(state_name - pp.Suppress(pp.Literal(","))) -
                        wait_type - pp.Opt(pp.Suppress(pp.Literal(",")) - state_name) -
                        pp.Opt(pp.CaselessKeyword("MAX").suppress() - wait_type) - pp.Suppress(pp.Literal(";"))).set_parse_action(WaitInstruction)

        opt_label = pp.Group(pp.Opt(identifier + pp.Suppress(pp.Literal(":"))))
        instruction <<= pp.MatchFirst((assignment, boolean, call, data, drscan, drstop, end_data, end_procedure, exit,
                                       export, for_, frequency, goto, if_, integer, irscan, irstop, next, note, pop,
                                       print_, procedure, push, state, trst, wait))
        statement = (opt_label + instruction).set_parse_action(LabelledInstruction)

        stapl_file = (pp.ZeroOrMore(note) - pp.ZeroOrMore(action) - pp.ZeroOrMore(statement) - crc - pp.StringEnd())

        stapl_file.ignore(comments)
        stapl_file.enable_packrat()

        logger.debug(f"Parsing stapl...")
        f = StaplFile(stapl_file.parse_string(f.read()))
        logger.debug(f"Stapl loaded")
        return f



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from bitarray import bitarray
from bitarray.util import int2ba, ba2int

from ..device import Pin

class MDIO:
    def __init__(self, ctl, MDC: Pin, MDIO: Pin, RESETn: Pin=None):
        self.ctl = ctl
        self.MDC = MDC
        self.MDIO = MDIO
        self.RESETn = RESETn

    async def init(self):
        self.MDC.output_enable(True)
        self.MDIO.output_enable(True)
        self.MDC.set_value(1)
        self.MDIO.set_value(1)
        if not self.RESETn is None:
            self.RESETn.output_enable(True)
            self.RESETn.set_value(0)
            await self.ctl.cycle()
            self.RESETn.set_value(1)
        await self.ctl.cycle()

    async def send_bits(self, bits: bitarray):
        # print("S", bits)
        self.MDIO.output_enable(True)
        for bit in bits:
            self.MDC.set_value(0)
            self.MDIO.set_value(bit)
            await self.ctl.cycle()
            self.MDC.set_value(1)
            await self.ctl.cycle()

    async def recv_bits(self, n):
        r = bitarray()
        self.MDIO.output_enable(False)
        for _ in range(n):
            self.MDC.set_value(0)
            await self.ctl.cycle()
            self.MDC.set_value(1)
            await self.ctl.cycle()
            r.append(self.MDIO.get_value())
        self.MDIO.output_enable(True)
        # print("R", r)
        return r

    async def read(self, phy_address, reg_address):
        await self.send_bits(bitarray("111111111111111111111111111111110110"))
        await self.send_bits(int2ba(phy_address, length=5))
        await self.send_bits(int2ba(reg_address, length=5))
        await self.recv_bits(2)
        return ba2int(await self.recv_bits(16))



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

from bitarray import bitarray

from .driver import Driver
from ..tap_controller import State
from ..device import Device

logger = logging.getLogger(__name__)

class Sim(Driver):
    def __init__(self, device: Device):
        Driver.__init__(self)
        self.device = device
        self.reset()

    def reset(self):
        self.state = State.TEST_LOGIC_RESET
        self.shift_ir = 0
        self.ir = 1
        self.dr_size = 1
        self.shift_dr = 0

    def transfer(self, tms, tdi):
        assert tms in (0, 1)
        assert tdi in (0, 1)

        tdo = 0
        next_state = self.state
        if self.state == State.TEST_LOGIC_RESET:
            if tms == 0:
                next_state = State.RUN_TEST_IDLE
        elif self.state == State.RUN_TEST_IDLE:
            if tms == 1:
                next_state = State.SELECT_DR_SCAN
        elif self.state == State.SELECT_DR_SCAN:
            if tms == 0:
                next_state = State.CAPTURE_DR
            elif tms == 1:
                next_state = State.SELECT_IR_SCAN
        elif self.state == State.CAPTURE_DR:
            if self.ir == self.device.opcodes['IDCODE']:
                logger.info("Shifting out ID code")
                self.shift_dr = self.device.idcode.to_bitarray()
                self.dr_size = 32
            elif self.ir == self.device.opcodes['BYPASS']:
                logger.info("Bypass")
                self.shift_dr = bitarray('0', endian='little')
                self.dr_size = 1
            elif self.ir == self.device.opcodes['SAMPLE']:
                logger.info("Sample")
                self.shift_dr = bitarray('0', endian='little')
                self.dr_size = len(self.device.cells)
            elif self.ir == self.device.opcodes['EXTEST']:
                logger.info("Extest")
                self.shift_dr = bitarray('0', endian='little')
                self.dr_size = len(self.device.cells)
            else:
                logger.warning(f"Unknown IR bin({self.ir})")
                self.shift_dr = bitarray('0', endian='little')
                self.dr_size = 1

            if tms == 0:
                next_state = State.SHIFT_DR
            elif tms == 1:
                next_state = State.EXIT1_DR
        elif self.state == State.SHIFT_DR:
            tdo = self.shift_dr[0]
            self.shift_dr = self.shift_dr[1:]
            self.shift_dr.append(tdi)
            if tms == 1:
                next_state = State.EXIT1_DR
        elif self.state == State.EXIT1_DR:
            if tms == 0:
                next_state = State.PAUSE_DR
            elif tms == 1:
                next_state = State.UPDATE_DR
        elif self.state == State.PAUSE_DR:
            if tms == 1:
                next_state = State.EXIT2_DR
        elif self.state == State.EXIT2_DR:
            if tms == 0:
                next_state = State.SHIFT_DR
            elif tms == 1:
                next_state = State.UPDATE_DR
        elif self.state == State.UPDATE_DR:
            if tms == 0:
                next_state = State.RUN_TEST_IDLE
            elif tms == 1:
                next_state = State.SELECT_DR_SCAN
        elif self.state == State.SELECT_IR_SCAN:
            if tms == 0:
                next_state = State.CAPTURE_IR
            elif tms == 1:
                next_state = State.TEST_LOGIC_RESET
        elif self.state == State.CAPTURE_IR:
            self.ir = bitarray('0' * self.device.irlen, endian='little') # TODO take INSTRUCTION_CAPTURE from BSDL
            self.shift_ir = self.ir
            if tms == 0:
                next_state = State.SHIFT_IR
            elif tms == 1:
                next_state = State.EXIT1_IR
        elif self.state == State.SHIFT_IR:
            tdo = self.shift_ir[0]
            self.shift_ir = self.shift_ir[1:]
            self.shift_ir.append(tdi)
            if tms == 1:
                next_state = State.EXIT1_IR
        elif self.state == State.EXIT1_IR:
            if tms == 0:
                next_state = State.PAUSE_IR
            elif tms == 1:
                next_state = State.UPDATE_IR
        elif self.state == State.PAUSE_IR:
            if tms == 1:
                next_state = State.EXIT2_IR
        elif self.state == State.EXIT2_IR:
            if tms == 0:
                next_state = State.SHIFT_IR
            elif tms == 1:
                next_state = State.UPDATE_IR
        elif self.state == State.UPDATE_IR:
            self.ir = self.shift_ir
            logger.info(f"Current instruction: {self.ir}")
            if tms == 0:
                next_state = State.RUN_TEST_IDLE
            elif tms == 1:
                next_state = State.SELECT_DR_SCAN
        else:
            assert False

        if self.state != next_state:
            logger.debug(f"State {self.state.name} => {next_state.name}")
            self.state = next_state

        return tdo

class SimChain(Driver):
    def __init__(self, devices: list[Sim] = []):
        self.devices = devices

    def transfer(self, tms, tdi):
        for device in self.devices:
            tdi = device.transfer(tms, tdi)
        return tdi


#!/usr/bin/env python3
import os
import logging

import ebyst

logger = logging.getLogger(__name__)

def get_all_bsdls(root):
    ret = []
    for (root, dirs, files) in os.walk(root):
        for dir in dirs:
            ret += get_all_bsdls(dir)
        for file in files:
            if os.path.splitext(file.lower())[-1] in (".bsm", ".bsd", ".bsdl"):
                ret.append(os.path.join(root, file))
    return ret

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)

    for fn in get_all_bsdls("bsdl"):
        logger.info(f"Parsing {fn}")
        dev = ebyst.Device.from_bsdl(fn)


#!/usr/bin/env python3
import asyncio
import logging

import ebyst

logger = logging.getLogger(__name__)

async def i2c_test(ctl):
    i2c = ebyst.interfaces.I2C(ctl, dev.pinmap['PB9A'], dev.pinmap['PB4B'])
    await i2c.init()

    dev_address = 0xa0 >> 1
    reg_address = 0x10
    data = 0xa5
    print(f"Writing {dev_address:02x}:{reg_address:02x} <= {data:02x}")
    await i2c.write(0xa0, 0x10, 0xa5)
    print(f"Reading {dev_address:02x}:{reg_address:02x} => ", end='')
    x = await i2c.read(0xa0, 0x10)
    print(f"{x:02x}")

    assert x == 0xa5

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    drv = ebyst.drivers.MPSSE(ebyst.drivers.MPSSE.list_devices()[0])
    dev = ebyst.Device.from_bsdl("bsdl/BSDLLCMXO2-256HCQFN32.BSM")
    ctl = ebyst.TapController(drv)
    ctl.detect_chain()
    ctl.add_device(dev)
    ctl.validate_chain()
    ctl.extest()

    try:
        asyncio.run(i2c_test(ctl))
    finally:
        ctl.reset()



#!/usr/bin/env python3
# Copyright (c) 2025 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os, sys
import logging
import argparse

import ebyst
from ebyst.stapl import StaplFile, StaplInterpreter

class ColorFormatter(logging.Formatter):
    """Custom formatter to add colors to log levels."""

    # Define color codes
    COLORS = {
        "DEBUG": "\033[0;90m",
        "INFO": "\033[1;97m",
        "WARNING": "\033[1;33m",
        "ERROR": "\033[1;31m",
        "CRITICAL": "\033[1;41m"
    }

    RESET = "\033[0;0m"

    def __init__(self, *args, color=True, **kwargs):
        logging.Formatter.__init__(self, *args, **kwargs)
        self.color = color

    def format(self, record):
        if self.color:
            log_color = self.COLORS.get(record.levelname, self.RESET)
            reset = self.RESET
        else:
            log_color = reset = ""
        record.levelname = f"{log_color}{record.levelname}{reset}"
        record.msg = f"{log_color}{record.msg}{reset}"
        return super().format(record)

logger = logging.getLogger(__name__)

def main():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    formatter = ColorFormatter("[%(asctime)s] [%(levelname)s] %(message)s", "%H:%M:%S", color=os.isatty(sys.stderr.fileno()))
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    for url in ebyst.drivers.MPSSE.list_devices([(0x1514, 0x2008)]):
        logger.info("Found device: %s", url)

    parser = argparse.ArgumentParser()
    parser.add_argument("device_url", type=str, help="FTDI URL")
    parser.add_argument("stapl_file", type=str, help="STAPL file")
    parser.add_argument("action", type=str, nargs='+', help="action(s) to perform")
    parser.add_argument("--optional", type=bool, default=False, help="Perform optional steps (default: false)")
    args = parser.parse_args()

    drv = ebyst.drivers.MPSSE(args.device_url)
    ctl = ebyst.TapController(drv)

    logger.info("Parsing stapl...")
    with open(args.stapl_file, "r") as f:
        stapl = StaplFile.parse(f)
    logger.info("Done")

    logger.info("Start running")
    interpreter = StaplInterpreter(ctl, stapl)
    for action in args.action:
        logger.info(f"Running action {action}")
        interpreter.run(action, optional=args.optional)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import logging
import asyncio
import time

import ebyst

logger = logging.getLogger(__name__)

async def blink(ctl):
    dev.pinmap['PB2D'].output_enable()
    while True:
        dev.pinmap['PB2D'].set_value(1)
        await ctl.cycle()
        time.sleep(0.5)

        dev.pinmap['PB2D'].set_value(0)
        await ctl.cycle()
        time.sleep(0.5)

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)

    drv = ebyst.drivers.MPSSE(ebyst.drivers.MPSSE.list_devices()[0])
    dev = ebyst.Device.from_bsdl("bsdl/BSDLLCMXO2-256HCQFN32.BSM")
    ctl = ebyst.TapController(drv)
    ctl.detect_chain()
    ctl.add_device(dev)
    ctl.validate_chain()
    ctl.extest()

    try:
        asyncio.run(blink(ctl))
    except KeyboardInterrupt:
        pass
    finally:
        ctl.reset()



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from enum import Enum, IntEnum
import logging
import asyncio
import time

from bitarray import bitarray

from .drivers.driver import Driver
from .device import Device
from .trace import Trace

logger = logging.getLogger(__name__)

MAX_IR_CHAIN_LENGTH = 255

class State(IntEnum):
    TEST_LOGIC_RESET = 0
    RUN_TEST_IDLE = 1
    SELECT_DR_SCAN = 10
    CAPTURE_DR = 11
    SHIFT_DR = 12
    EXIT1_DR = 13
    PAUSE_DR = 14
    EXIT2_DR = 15
    UPDATE_DR = 16
    SELECT_IR_SCAN = 20
    CAPTURE_IR = 21
    SHIFT_IR = 22
    EXIT1_IR = 23
    PAUSE_IR = 24
    EXIT2_IR = 25
    UPDATE_IR = 26

class Opcode(Enum):
    BYPASS = "BYPASS"
    IDCODE = "IDCODE"
    SAMPLE = "SAMPLE"
    PRELOAD = "PRELOAD"
    EXTEST = "EXTEST"
    EXTEST_PULSE = "EXTEST_PULSE"

class TapController:
    class Chain(list):
        """Chain of devices, 0 is close to TDI"""
        def __init__(self, *args, **kwargs):
            list.__init__(self, *args, **kwargs)
            self.validated = False

        @property
        def brlen(self):
            r = 0
            for dev in self:
                r += len(dev.cells)
            return r

        def generate_ir(self, instruction):
            tdi_str = bitarray(endian='little')
            for dev in self:
                try:
                    tdi_str = dev.opcodes[instruction.value] + tdi_str
                except KeyError:
                    raise Exception(f"Instruction {instruction} not supported by all devices in chain")
            return tdi_str

        def reset(self):
            for dev in self:
                dev.reset()

        def generate_br(self):
            tdi_str = bitarray(endian='little')
            for dev in self:
                tdi_str = dev.generate_br() + tdi_str
            return tdi_str

        def update_br(self, br):
            o = 0
            for dev in self:
                l = len(dev.cells)
                dev.update_br(br[o:o+l])
                o += l

    def __init__(self, driver: Driver, max_freq=None, no_parallel=False):
        self.driver = driver
        self.driver.set_freq(100e3)
        self.driver.reset()
        self.state = State.TEST_LOGIC_RESET
        self.chain = TapController.Chain()
        self.in_extest = False
        self.no_parallel = no_parallel
        self.cycle_counter = 0
        self.traces = []
        self.max_freq = max_freq

    def reset(self):
        self.driver.reset()
        self.state = State.TEST_LOGIC_RESET

    def load_instruction(self, instruction: Opcode):
        if not self.chain.validated: raise Exception("Chain not validated")
        tdi_str = self.chain.generate_ir(instruction)
        logger.debug(f"Loading instruction {instruction}")
        self._goto(State.SHIFT_IR)
        self.driver.transmit_tdi_str(tdi_str, first_tms=0 if len(tdi_str) > 1 else 1, last_tms=1)
        self.state = State.EXIT1_IR
        self._goto(State.UPDATE_IR)
        self.in_extest = False

    def read_register(self, n: int):
        if not self.chain.validated: raise Exception("Chain not validated")
        self._goto(State.SHIFT_DR)
        tdo = self.driver.receive_tdo_str(n, first_tms=0 if n > 1 else 1, last_tms=1)
        self.state = State.EXIT1_DR
        self._goto(State.UPDATE_DR)
        return tdo

    def write_register(self, tdi: bitarray):
        if not self.chain.validated: raise Exception("Chain not validated")
        self._goto(State.SHIFT_DR)
        self.driver.transmit_tdi_str(tdi, first_tms=0 if len(tdi) > 1 else 1, last_tms=1)
        self.state = State.EXIT1_DR
        self._goto(State.UPDATE_DR)

    def read_write_register(self, tdi: bitarray):
        if not self.chain.validated: raise Exception("Chain not validated")
        self._goto(State.SHIFT_DR)
        tdo = self.driver.transfer_tdi_tdo_str(tdi, first_tms=0 if len(tdi) > 1 else 1, last_tms=1)
        self.state = State.EXIT1_DR
        self._goto(State.UPDATE_DR)
        return tdo

    def detect_chain(self):
        """Detect chain length (nr of devices and total instruction register length"""
        try:
            self._goto(State.SHIFT_IR)
            # load IR with 0
            self.driver.transmit_tdi_str(bitarray('0', endian='little') * MAX_IR_CHAIN_LENGTH)
            if self.driver.transfer(0, 0) != 0:
                raise Exception("Chain detection failed: TDO stuck at 1")
            # load IR with 1, find rising edge position
            irlen = 0
            while self.driver.transfer(0, 1) == 0:
                irlen += 1
                if irlen >= MAX_IR_CHAIN_LENGTH:
                    raise Exception("Chain detection failed: TDO stuck at 0")
            # IR is now all 1's => BYPASS
            self._goto(State.UPDATE_IR, 1)
            self._goto(State.SHIFT_DR)
            # load BYPASS with 0
            self.driver.transmit_tdi_str(bitarray('0', endian='little') * MAX_IR_CHAIN_LENGTH)
            if self.driver.transfer(0, 0) != 0:
                raise Exception("Chain detection failed: TDO stuck at 1")
            # load BYPASS with 1, find rising edge position
            drlen = 0
            while self.driver.transfer(0, 1) == 0:
                drlen += 1
                if drlen >= MAX_IR_CHAIN_LENGTH:
                    raise Exception("Chain detection failed: TDO stuck at 0")
        finally:
            self._goto(State.TEST_LOGIC_RESET)

        logger.info(f"Found {drlen} device(s) with a total IR chain length of {irlen}")
        return (drlen, irlen)

    def add_device(self, device: Device):
        """Add device to chain"""
        self.chain.append(device)
        self.chain.validated = False

    def validate_chain(self):
        """Validate configured chain"""
        drlen, irlen = self.detect_chain()
        if len(self.chain) != 1:
            logger.warning("Multiple devices in chain are not tested")
        if drlen != len(self.chain):
            raise Exception(f"Incorrect nr of devices in chain ({drlen} detected)")
        if irlen != sum(dev.irlen for dev in self.chain):
            raise Exception(f"Incorrect total ir length ({irlen} detected)")

        max_freq = self.max_freq
        for dev in self.chain:
            if max_freq is None or dev.max_freq < max_freq:
                max_freq = dev.max_freq
        if not max_freq is None:
            self.driver.set_freq(max_freq)

        try:
            self.chain.validated = True
            self.load_instruction(Opcode.IDCODE)
            idcode = self.read_register(32*drlen)
            for i, dev in enumerate(self.chain):
                if idcode[len(idcode)-i*32-32:len(idcode)-i*32] != dev.idcode:
                    raise Exception(f"IDCode doesn't match for device {i} {idcode}<=>{dev.idcode}")

            self.load_instruction(Opcode.SAMPLE)
            br = self.read_register(self.chain.brlen)
            self.chain.update_br(br)
        except:
            self.chain.validated = False
            raise
        finally:
            self._goto(State.RUN_TEST_IDLE)

    def set_frequency(self, frequency):
        self.driver.set_freq(min(frequency, self.max_freq or frequency))

    def extest(self):
        self.in_extest = False
        self.load_instruction(Opcode.SAMPLE)
        br = self.chain.generate_br()
        br = self.read_write_register(br)
        self.chain.update_br(br)
        self.load_instruction(Opcode.EXTEST)
        self.in_extest = True

    def extest_pulse(self):
        self.load_instruction(Opcode.EXTEST_PULSE)
        self.chain.reset()

    async def cycle(self):
        """Cycle the boundary scan register when in extest() mode; updates the output pins,
            and samples the input pins
        """
        # Force a reschedule before cycling BR, this allows all tasks to share a BR cycle.
        # If we are awaken again, and no other task performed the cycle; we execute it.
        cycle_counter = self.cycle_counter
        if not self.no_parallel: await asyncio.sleep(0)
        if cycle_counter == self.cycle_counter:
            # nobody else performed the cycle while we where sleeping => we do it
            br = self.chain.generate_br()
            br = self.read_write_register(br)
            self.chain.update_br(br)
            for trace in self.traces: trace.snapshot()
            self.cycle_counter += 1

    def trace(self, fn, **pins):
        self.traces.append(Trace(fn, **pins))

    def enter_state(self, state: State):
        self._goto(state)

    def _goto(self, target_state: State, tdi=0):
        state = self.state
        if state == target_state: return
        logger.debug(f"Going from {state.name} to {target_state.name}")
        tms = bitarray(endian='little')
        while state != target_state:
            if state == State.TEST_LOGIC_RESET:
                tms.append(0)
                state = State.RUN_TEST_IDLE
            elif state == State.RUN_TEST_IDLE:
                tms.append(1)
                state = State.SELECT_DR_SCAN
            elif state == State.SELECT_DR_SCAN:
                if target_state > State.SELECT_DR_SCAN and target_state <= State.UPDATE_DR:
                    tms.append(0)
                    state = State.CAPTURE_DR
                else:
                    tms.append(1)
                    state = State.SELECT_IR_SCAN
            elif state == State.CAPTURE_DR:
                if target_state == State.SHIFT_DR:
                    tms.append(0)
                    state = State.SHIFT_DR
                else:
                    tms.append(1)
                    state = State.EXIT1_DR
            elif state == State.SHIFT_DR:
                tms.append(1)
                state = State.EXIT1_DR
            elif state == State.EXIT1_DR:
                if target_state in (State.PAUSE_DR, State.EXIT2_DR, State.SHIFT_DR):
                    tms.append(0)
                    state = State.PAUSE_DR
                else:
                    tms.append(1)
                    state = State.UPDATE_DR
            elif state == State.PAUSE_DR:
                tms.append(1)
                state = State.EXIT2_DR
            elif state == State.EXIT2_DR:
                if target_state in (State.SHIFT_DR, State.EXIT1_DR, State.PAUSE_DR):
                    tms.append(0)
                    state = State.SHIFT_DR
                else:
                    tms.append(1)
                    state = State.UPDATE_DR
            elif state == State.UPDATE_DR:
                if target_state == State.RUN_TEST_IDLE:
                    tms.append(0)
                    state = State.RUN_TEST_IDLE
                else:
                    tms.append(1)
                    state = State.SELECT_DR_SCAN
            elif state == State.SELECT_IR_SCAN:
                if target_state > State.SELECT_IR_SCAN and target_state <= State.UPDATE_IR:
                    tms.append(0)
                    state = State.CAPTURE_IR
                else:
                    tms.append(1)
                    state = State.TEST_LOGIC_RESET
            elif state == State.CAPTURE_IR:
                if target_state == State.SHIFT_IR:
                    tms.append(0)
                    state = State.SHIFT_IR
                else:
                    tms.append(1)
                    state = State.EXIT1_IR
            elif state == State.SHIFT_IR:
                tms.append(1)
                state = State.EXIT1_IR
            elif state == State.EXIT1_IR:
                if target_state in (State.PAUSE_IR, State.EXIT2_IR, State.SHIFT_IR):
                    tms.append(0)
                    state = State.PAUSE_IR
                else:
                    tms.append(1)
                    state = State.UPDATE_IR
            elif state == State.PAUSE_IR:
                tms.append(1)
                state = State.EXIT2_IR
            elif state == State.EXIT2_IR:
                if target_state in (State.SHIFT_IR, State.EXIT1_IR, State.PAUSE_IR):
                    tms.append(0)
                    state = State.SHIFT_IR
                else:
                    tms.append(1)
                    state = State.UPDATE_IR
            elif state == State.UPDATE_IR:
                if target_state == State.RUN_TEST_IDLE:
                    tms.append(0)
                    state = State.RUN_TEST_IDLE
                else:
                    tms.append(1)
                    state = State.SELECT_DR_SCAN
            else:
                assert False

        logger.debug(f"TMS string: {tms}")

        self.driver.transmit_tms_str(tms, tdi)
        self.state = state

    def export(self, key, value):
        print(f"EXPORT {key}={value}")

    def wait(self, cycles: int, usec: int=0):
        """Wait for until both (tck-)cycles and usec are satisfied"""
        if self.state in (State.RUN_TEST_IDLE, State.PAUSE_DR, State.PAUSE_IR):
            self.driver.transmit_tms_str(bitarray("0" * cycles, 'little'))
        elif self.state in (State.TEST_LOGIC_RESET, ):
            self.driver.transmit_tms_str(bitarray("1" * cycles, 'little'))
        else:
            raise Exception("{self.state} is not a wait state")
        if usec > 500: time.sleep(usec * 1e-6)

    def ir_scan(self, ir: bitarray, end_state: State | None=None):
        if ir.endian != 'little': raise ValueError("ir must be little endian bitarray")
        if end_state is None:
            logger.debug(f"IR scan {ir}")
        else:
            logger.debug(f"IR scan {ir} exit to {end_state.name}")
        self._goto(State.SHIFT_IR)
        ret = self.driver.transfer_tdi_tdo_str(ir, first_tms=0 if len(ir) > 1 else 1, last_tms=1)
        self.state = State.EXIT1_IR
        if not end_state is None: self._goto(end_state)
        self.in_extest = False
        return ret

    def dr_scan(self, dr: bitarray, end_state: State | None=None):
        if dr.endian != 'little': raise ValueError("dr must be little endian bitarray")
        if end_state is None:
            logger.debug(f"DR scan {dr}")
        else:
            logger.debug(f"DR scan {dr} exit to {end_state.name}")
        self._goto(State.SHIFT_DR)
        ret = self.driver.transfer_tdi_tdo_str(dr, first_tms=0 if len(dr) > 1 else 1, last_tms=1)
        self.state = State.EXIT1_DR
        if not end_state is None: self._goto(end_state)
        self.in_extest = False
        return ret



#!/usr/bin/env python3
# Copyright (c) 2025 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import math
from bitarray.util import ba2int, hex2ba
from bitarray import bitarray
from binascii import hexlify

def _get_6bits(compressed: str):
    for c in compressed:
        c = ord(c)
        if c >= 0x30 and c <= 0x39:
            yield c - 0x30
        elif c >= 0x41 and c <= 0x5a:
            yield c - 0x41 + 10
        elif c >= 0x61 and c <= 0x7a:
            yield c - 0x61 + 36
        elif c == 0x5f:
            yield 0x3e
        elif c == 0x40:
            yield 0x3f
        elif c in (0x08, 0x0a, 0x0d, 0x20):
            pass
        else:
            raise ValueError(f"Invalid character in compressed stream (0x{c:02x})")

def _get_bytes(compressed: str):
    it = _get_6bits(compressed)
    try:
        while True:
            a = next(it)
            b = next(it)
            yield a | ((b & 0x03) << 6)
            c = next(it)
            yield ((b & 0x3c) >> 2) | ((c & 0x0f) << 4)
            d = next(it)
            yield ((c & 0x30) >> 4) | (d << 2)
    except StopIteration:
        pass

def decompress(compressed: str):
    ba = bitarray(endian='little', buffer=bytes(_get_bytes(compressed)))
    length = ba2int(ba[:32])
    ret = bytearray(length)
    ioffset = 32
    ooffset = 0
    while ooffset < length:
        if ba[ioffset] == 0:
            ioffset += 1
            for _ in range(3):
                ret[ooffset:ooffset+1] = bytes([ba2int(ba[ioffset:ioffset+8])])
                ioffset += 8
                ooffset += 1
        else:
            ioffset += 1
            bits = min(math.ceil(math.log2(ooffset)), 13)
            repeat_offset = ba2int(ba[ioffset:ioffset+bits])
            ioffset += bits
            repeat_length = ba2int(ba[ioffset:ioffset+8])
            ioffset += 8
            ret[ooffset:ooffset+repeat_length] = ret[ooffset-repeat_offset:ooffset-repeat_offset+repeat_length]
            ooffset += repeat_length

    return ret

if __name__ == "__main__":
    assert decompress("O00008Cn63PbPMRWpGBDgj6RV60") == b"abcdefabcdefghijkldefabc"

    v = bitarray("101101111", endian='little')
    v.reverse()
    print(f"BIN: {v}")

    v = bitarray(hex2ba("16F", endian='big'), 'little')
    v.reverse()
    print(f"HEX: {v}")

    x = decompress("30000uj000")
    v = bitarray(endian='little')
    v.frombytes(x)
    print(f"ACA: {v}")



#!/usr/bin/env python3
import asyncio
import logging

import ebyst

logger = logging.getLogger(__name__)

async def lb_test(ctl):
    await ctl.cycle()

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    dev1 = ebyst.Device.from_bsdl("bsdl/BSDLLCMXO2-256HCQFN32.BSM")
    dev2 = ebyst.Device.from_bsdl("bsdl/MPF100Tfcg484.bsdl")
    dev3 = ebyst.Device.from_bsdl("bsdl/XA3S400_FG456.bsdl")
    sim1 = ebyst.drivers.Sim(dev1)
    sim2 = ebyst.drivers.Sim(dev2)
    sim3 = ebyst.drivers.Sim(dev3)
    sim_chain = ebyst.drivers.SimChain([sim1, sim2, sim3])
    ctl = ebyst.TapController(sim_chain)
    ctl.detect_chain()
    ctl.add_device(dev1)
    ctl.add_device(dev2)
    ctl.add_device(dev3)
    ctl.validate_chain()
    ctl.extest()

    try:
        asyncio.run(lb_test(ctl))
    finally:
        ctl.reset()



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from bitarray import bitarray

from ..device import Pin

class SPI:
    def __init__(self, ctl, SCK: Pin, SSn: Pin, MOSI: Pin, MISO: Pin):
        self.ctl = ctl
        self.SCK = SCK
        self.SSn = SSn
        self.MOSI = MOSI
        self.MISO = MISO

    async def init(self):
        self.SCK.output_enable(True)
        self.SSn.output_enable(True)
        self.MOSI.output_enable(True)
        self.MISO.output_enable(False)
        self.SCK.set_value(0)
        self.SSn.set_value(1)
        await self.ctl.cycle()

    async def transfer(self, bits: bitarray):
        r = bitarray()
        self.SSn.set_value(0)

        for bit in bits:
            self.MOSI.set_value(bit)
            await self.ctl.cycle()
            self.SCK.set_value(1)
            await self.ctl.cycle()
            r.append(self.MISO.get_value())
            self.SCK.set_value(0)

        self.SSn.set_value(1)
        await self.ctl.cycle()

        return r


# Copyright (c) 2025 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
class StaplError(Exception):
    def __init__(self, message, pc=None):
        self.pc = pc
        self.message = message

    def __str__(self):
        if not self.pc is None:
            r = f"{self.pc}: {self.message}"
        else:
            r = self.message
        return r

class VariableNotDefined(StaplError):
    def __init__(self, var):
        StaplError.__init__(self, f"Variable {var} not defined")
        self.var = var

class InvalidState(StaplError):
    def __init__(self, state):
        StaplError.__init__(self, f"Invalid state {state}")
        self.state = state

class LabelNotDefined(StaplError):
    def __init__(self, label):
        StaplError.__init__(self, f"Label {label} not defined")
        self.label = label

class StaplValueError(StaplError):
    pass



#!/usr/bin/env python3
import logging

import ebyst

logger = logging.getLogger(__name__)

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)

    drv = ebyst.drivers.MPSSE("ftdi://ftdi:2232:251633009FEC/1")
    drv.test()
